import math
from qibo.backends import abstract, einsum_utils
from qibo.config import raise_error, log


class NumpyBackend(abstract.AbstractBackend):

    description = "Uses `np.einsum` to apply gates to states via matrix " \
                  "multiplication."

    TEST_REGRESSIONS = {
        "test_measurementresult_apply_bitflips": [
                [0, 0, 0, 0, 2, 3, 0, 0, 0, 0],
                [0, 0, 0, 0, 2, 3, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 2, 0, 0, 0, 0, 0]
            ],
        "test_probabilistic_measurement": {0: 249, 1: 231, 2: 253, 3: 267},
        "test_unbalanced_probabilistic_measurement": {0: 171, 1: 148, 2: 161, 3: 520},
        "test_post_measurement_bitflips_on_circuit": [
                {5: 30}, {5: 18, 4: 5, 7: 4, 1: 2, 6: 1},
                {4: 8, 2: 6, 5: 5, 1: 3, 3: 3, 6: 2, 7: 2, 0: 1}
            ],
    }

    def __init__(self):
        super().__init__()
        import numpy as np
        self.backend = np
        self.name = "numpy"
        self.np = np

        self.numeric_types = (np.int, np.float, np.complex, np.int32,
                              np.int64, np.float32, np.float64,
                              np.complex64, np.complex128)
        self.tensor_types = (np.ndarray,)
        self.native_types = (np.ndarray,)
        self.Tensor = np.ndarray
        self.random = np.random
        self.newaxis = np.newaxis
        self.oom_error = MemoryError
        self.optimization = None
        self.cpu_devices = ["/CPU:0"]
        self.gpu_devices = []
        self.default_device = self.cpu_devices[0]

    def set_device(self, name):
        log.warning("Numpy does not support device placement. "
                    "Aborting device change.")

    def set_threads(self, nthreads):
        log.warning("Numpy backend supports only single-thread execution. "
                    "Cannot change the number of threads.")
        abstract.AbstractBackend.set_threads(self, nthreads)

    def to_numpy(self, x):
        return x

    def to_complex(self, re, img): # pragma: no cover
        return re + 1j * img

    def cast(self, x, dtype='DTYPECPX'):
        if isinstance(dtype, str):
            dtype = self.dtypes(dtype)
        if isinstance(x, self.backend.ndarray):
            return x.astype(dtype)
        return self.backend.array(x, dtype=dtype)

    def diag(self, x, dtype='DTYPECPX'):
        if isinstance(dtype, str):
            dtype = self.dtypes(dtype)
        return self.backend.diag(x).astype(dtype)

    def reshape(self, x, shape):
        return self.backend.reshape(x, shape)

    def stack(self, x, axis=0):
        return self.backend.stack(x, axis=axis)

    def concatenate(self, x, axis=0):
        return self.backend.concatenate(x, axis=axis)

    def expand_dims(self, x, axis):
        return self.backend.expand_dims(x, axis)

    def copy(self, x):
        return self.backend.copy(x)

    def range(self, start, finish, step, dtype=None):
        if isinstance(dtype, str):
            dtype = self.dtypes(dtype)
        return self.backend.arange(start, finish, step, dtype=dtype)

    def eye(self, shape, dtype='DTYPECPX'):
        if isinstance(dtype, str):
            dtype = self.dtypes(dtype)
        return self.backend.eye(shape, dtype=dtype)

    def zeros(self, shape, dtype='DTYPECPX'):
        if isinstance(dtype, str):
            dtype = self.dtypes(dtype)
        return self.backend.zeros(shape, dtype=dtype)

    def ones(self, shape, dtype='DTYPECPX'):
        if isinstance(dtype, str):
            dtype = self.dtypes(dtype)
        return self.backend.ones(shape, dtype=dtype)

    def zeros_like(self, x):
        return self.backend.zeros_like(x)

    def ones_like(self, x):
        return self.backend.ones_like(x)

    def real(self, x):
        return self.backend.real(x)

    def imag(self, x):
        return self.backend.imag(x)

    def conj(self, x):
        return self.backend.conj(x)

    def mod(self, x, y):
        return self.backend.mod(x, y)

    def right_shift(self, x, y):
        return self.backend.right_shift(x, y)

    def exp(self, x):
        return self.backend.exp(x)

    def sin(self, x):
        return self.backend.sin(x)

    def cos(self, x):
        return self.backend.cos(x)

    def pow(self, base, exponent):
        return base ** exponent

    def square(self, x):
        return x ** 2

    def sqrt(self, x):
        return self.backend.sqrt(x)

    def log(self, x):
        return self.backend.log(x)

    def abs(self, x):
        return self.backend.abs(x)

    def trace(self, x):
        return self.backend.trace(x)

    def expm(self, x):
        from scipy import linalg
        return linalg.expm(x)

    def sum(self, x, axis=None):
        return self.backend.sum(x, axis=axis)

    def matmul(self, x, y):
        return self.backend.matmul(x, y)

    def outer(self, x, y):
        return self.backend.outer(x, y)

    def kron(self, x, y):
        return self.backend.kron(x, y)

    def einsum(self, *args):
        return self.backend.einsum(*args)

    def tensordot(self, x, y, axes=None):
        return self.backend.tensordot(x, y, axes=axes)

    def transpose(self, x, axes=None):
        return self.backend.transpose(x, axes)

    def inv(self, x):
        return self.backend.linalg.inv(x)

    def eigh(self, x):
        return self.backend.linalg.eigh(x)

    def eigvalsh(self, x):
        return self.backend.linalg.eigvalsh(x)

    def less(self, x, y):
        return self.backend.less(x, y)

    def array_equal(self, x, y):
        return self.np.array_equal(x, y)

    def unique(self, x, return_counts=False):
        # Uses numpy backend always (even on Tensorflow)
        return self.np.unique(x, return_counts=return_counts)

    def squeeze(self, x, axis=None):
        return self.backend.squeeze(x, axis=axis)

    def gather(self, x, indices=None, condition=None, axis=0):
        if indices is None:
            if condition is None:
                raise_error(ValueError, "Gather call requires either indices "
                                        "or condition.")
            indices = condition
        if axis < 0:
            axis += len(x.shape)
        idx = axis * (slice(None),) + (indices,)
        return x[idx]

    def gather_nd(self, x, indices):
        return x[tuple(indices)]

    def initial_state(self, nqubits, is_matrix=False):
        if is_matrix:
            state = self.zeros(2 * (2 ** nqubits,))
            state[0, 0] = 1
        else:
            state = self.zeros((2 ** nqubits,))
            state[0] = 1
        return state

    def random_uniform(self, shape, dtype='DTYPE'):
        return self.backend.random.random(shape).astype(self.dtypes(dtype))

    def sample_shots(self, probs, nshots):
        return self.random.choice(range(len(probs)), size=nshots, p=probs)

    def sample_frequencies(self, probs, nshots):
        from qibo.config import SHOT_BATCH_SIZE
        def update_frequencies(nsamples, frequencies):
            samples = self.random.choice(
                range(len(probs)), size=nsamples, p=probs/self.backend.sum(probs))
            res, counts = self.backend.unique(samples, return_counts=True)
            frequencies[res] += counts
            return frequencies

        frequencies = self.zeros(int(probs.shape[0]), dtype=self.dtypes('DTYPEINT'))
        for _ in range(nshots // SHOT_BATCH_SIZE):
            frequencies = update_frequencies(SHOT_BATCH_SIZE, frequencies)
        frequencies = update_frequencies(nshots % SHOT_BATCH_SIZE, frequencies)
        return frequencies

    def compile(self, func):
        return func

    class DummyModule:

        def __enter__(self, *args):
            pass

        def __exit__(self, *args):
            pass

    def device(self, device_name):
        return self.DummyModule()

    def set_seed(self, seed):
        self.backend.random.seed(seed)

    def create_einsum_cache(self, qubits, nqubits, ncontrol=None):
        return einsum_utils.EinsumCache(qubits, nqubits, ncontrol)

    def einsum_call(self, cache, state, matrix):
        return self.einsum(cache, state, matrix)

    class GateCache:
        pass

    def create_gate_cache(self, gate):
        cache = self.GateCache()
        s = 1 + gate.density_matrix
        cache.tensor_shape = self.cast(s * gate.nqubits * (2,), dtype='DTYPEINT')
        cache.flat_shape = self.cast(s * (2 ** gate.nqubits,), dtype='DTYPEINT')
        if gate.is_controlled_by:
            cache.control_cache = einsum_utils.ControlCache(gate)
            nactive = gate.nqubits - len(gate.control_qubits)
            targets = cache.control_cache.targets
            ncontrol = len(gate.control_qubits)
            cache.calculation_cache = self.create_einsum_cache(
                targets, nactive, ncontrol)
        else:
            cache.calculation_cache = self.create_einsum_cache(gate.qubits, gate.nqubits)
        return cache

    def reshape_matrix(self, matrix):
        """Reshapes the gate matrix to the tensor shape.

        Helper method for gate calls.
        """
        rank = int(math.log2(int(matrix.shape[0])))
        return self.reshape(matrix, 2 * rank * (2,))

    def _state_vector_call(self, gate, state):
        state = self.reshape(state, gate.cache.tensor_shape)
        matrix = self.reshape_matrix(gate.native_op_matrix)
        if gate.is_controlled_by:
            ncontrol = len(gate.control_qubits)
            nactive = gate.nqubits - ncontrol
            state = self.transpose(state, gate.cache.control_cache.order(False))
            # Apply `einsum` only to the part of the state where all controls
            # are active. This should be `state[-1]`
            state = self.reshape(state, (2 ** ncontrol,) + nactive * (2,))
            updates = self.einsum_call(gate.cache.calculation_cache.vector, state[-1], matrix)
            # Concatenate the updated part of the state `updates` with the
            # part of of the state that remained unaffected `state[:-1]`.
            state = self.concatenate([state[:-1], updates[self.newaxis]], axis=0)
            state = self.reshape(state, gate.nqubits * (2,))
            # Put qubit indices back to their proper places
            state = self.transpose(state, gate.cache.control_cache.reverse(False))
        else:
            einsum_str = gate.cache.calculation_cache.vector
            state = self.einsum_call(einsum_str, state, matrix)
        return self.reshape(state, gate.cache.flat_shape)

    def state_vector_matrix_call(self, gate, state):
        return self._state_vector_call(gate, state)

    def _density_matrix_call(self, gate, state):
        state = self.reshape(state, gate.cache.tensor_shape)
        matrix = self.reshape_matrix(gate.native_op_matrix)
        matrixc = self.conj(matrix)
        if gate.is_controlled_by:
            ncontrol = len(gate.control_qubits)
            nactive = gate.nqubits - ncontrol
            n = 2 ** ncontrol
            state = self.transpose(state, gate.cache.control_cache.order(True))
            state = self.reshape(state, 2 * (n,) + 2 * nactive * (2,))
            state01 = self.gather(state, indices=range(n - 1), axis=0)
            state01 = self.squeeze(self.gather(state01, indices=[n - 1], axis=1), axis=1)
            state01 = self.einsum_call(gate.cache.calculation_cache.right0, state01, matrixc)
            state10 = self.gather(state, indices=range(n - 1), axis=1)
            state10 = self.squeeze(self.gather(state10, indices=[n - 1], axis=0), axis=0)
            state10 = self.einsum_call(gate.cache.calculation_cache.left0,
                                       state10, matrix)

            state11 = self.squeeze(self.gather(state, indices=[n - 1], axis=0), axis=0)
            state11 = self.squeeze(self.gather(state11, indices=[n - 1], axis=0), axis=0)
            state11 = self.einsum_call(gate.cache.calculation_cache.right, state11, matrixc)
            state11 = self.einsum_call(gate.cache.calculation_cache.left, state11, matrix)

            state00 = self.gather(state, indices=range(n - 1), axis=0)
            state00 = self.gather(state00, indices=range(n - 1), axis=1)
            state01 = self.concatenate([state00, state01[:, self.newaxis]], axis=1)
            state10 = self.concatenate([state10, state11[self.newaxis]], axis=0)
            state = self.concatenate([state01, state10[self.newaxis]], axis=0)
            state = self.reshape(state, 2 * gate.nqubits * (2,))
            state = self.transpose(state, gate.cache.control_cache.reverse(True))
        else:
            state = self.einsum_call(gate.cache.calculation_cache.right, state, matrixc)
            state = self.einsum_call(gate.cache.calculation_cache.left, state, matrix)
        return self.reshape(state, gate.cache.flat_shape)

    def density_matrix_matrix_call(self, gate, state):
        return self._density_matrix_call(gate, state)

    def _density_matrix_half_call(self, gate, state):
        if gate.is_controlled_by: # pragma: no cover
            raise_error(NotImplementedError, "Gate density matrix half call is "
                                             "not implemented for ``controlled_by``"
                                             "gates.")
        matrix = self.reshape_matrix(gate.native_op_matrix)
        state = self.reshape(state, gate.cache.tensor_shape)
        state = self.einsum_call(gate.cache.calculation_cache.left, state, matrix)
        return self.reshape(state, gate.cache.flat_shape)

    def density_matrix_half_matrix_call(self, gate, state):
        return self._density_matrix_half_call(gate, state)

    def _append_zeros(self, state, qubits, results):
        """Helper method for `state_vector_collapse` and `density_matrix_collapse`."""
        for q, r in zip(qubits, results):
            state = self.expand_dims(state, axis=q)
            if r:
                state = self.concatenate([self.zeros_like(state), state], axis=q)
            else:
                state = self.concatenate([state, self.zeros_like(state)], axis=q)
        return state

    def state_vector_collapse(self, gate, state, result):
        state = self.reshape(state, gate.cache.tensor_shape)
        substate = self.gather_nd(self.transpose(state, gate.cache.order), result)
        norm = self.sum(self.square(self.abs(substate)))
        state = substate / self.cast(self.sqrt(norm), dtype=state.dtype)
        state = self._append_zeros(state, sorted(gate.target_qubits), result)
        return self.reshape(state, gate.cache.flat_shape)

    def density_matrix_collapse(self, gate, state, result):
        density_matrix_result = 2 * result
        sorted_qubits = sorted(gate.target_qubits)
        sorted_qubits = sorted_qubits + [q + gate.nqubits for q in sorted_qubits]
        state = self.reshape(state, gate.cache.tensor_shape)
        substate = self.gather_nd(self.transpose(state, gate.cache.order),
                                  density_matrix_result)
        n = 2 ** (len(tuple(substate.shape)) // 2)
        norm = self.trace(self.reshape(substate, (n, n)))
        state = substate / norm
        state = self._append_zeros(state, sorted_qubits, density_matrix_result)
        return self.reshape(state, gate.cache.flat_shape)

    def on_cpu(self): # pragma: no cover
        return self.device()

    def cpu_tensor(self, x, dtype=None):
        if dtype is None:
            dtype = x.dtype
        return self.np.asarray(x, dtype=dtype)

    def cpu_cast(self, x, dtype='DTYPECPX'):
        dtype = self._dtypes.get(dtype)
        return self.np.array(x, dtype=dtype)

    def cpu_assign(self, state, i, piece):
        state.pieces[i] = self.to_numpy(piece)

    def transpose_state(self, pieces, state, nqubits, order):
        pieces = self.reshape(self.backend.stack(pieces), nqubits * (2,))
        return self.reshape(self.transpose(pieces, order), state.shape)

    def assert_allclose(self, value, target, rtol=1e-7, atol=0.0):
        self.np.testing.assert_allclose(value, target, rtol=rtol, atol=atol)



"""Tests methods defined in `qibo/abstractions/gates.py` and `qibo/abstractions/abstract_gates.py`."""
import pytest
from qibo.abstractions import gates


@pytest.mark.parametrize("gatename", ["H", "X", "Y", "Z", "I", "Align"])
def test_one_qubit_gates_init(gatename):
    gate = getattr(gates, gatename)(0)
    assert gate.target_qubits == (0,)


@pytest.mark.parametrize("controls,instance",
                         [((1,), "CNOT"), ((1, 2), "TOFFOLI"),
                          ((1, 2, 4), "X")])
def test_x_controlled_by(controls, instance):
    gate = gates.X(0).controlled_by(*controls)
    assert gate.target_qubits == (0,)
    assert gate.control_qubits == controls
    assert isinstance(gate, getattr(gates, instance))


def test_x_decompose_with_few_controls():
    """Check ``X`` decomposition with less than three controls."""
    gate = gates.X(0)
    decomp = gate.decompose(1, 2)
    assert len(decomp) == 1
    assert isinstance(decomp[0], gates.X)


@pytest.mark.parametrize("use_toffolis", [True, False])
def test_x_decomposition_errors(use_toffolis):
    """Check ``X`` decomposition errors."""
    gate = gates.X(0).controlled_by(1, 2, 3, 4)
    with pytest.raises(ValueError):
        decomp = gate.decompose(2, 3, use_toffolis=use_toffolis)
    gate.nqubits = 6
    with pytest.raises(ValueError):
        decomp = gate.decompose(5, 6, use_toffolis=use_toffolis)


@pytest.mark.parametrize("controls,instance", [((1,), "CZ"), ((1, 2), "Z")])
def test_z_controlled_by(controls, instance):
    gate = gates.Z(0).controlled_by(*controls)
    assert gate.target_qubits == (0,)
    assert gate.control_qubits == controls
    assert isinstance(gate, getattr(gates, instance))


@pytest.mark.parametrize("targets,p0,p1",
                         [((0,), None, None),
                          ((0, 1, 2), None, None),
                          ((0, 3, 2), 0.2, 0.1)])
def test_measurement_init(targets, p0, p1):
    # also tests `_get_bitflip_map`
    gate = gates.M(*targets, p0=p0, p1=p1)
    assert gate.target_qubits == targets
    p0map = {q: 0 if p0 is None else p0 for q in targets}
    p1map = {q: 0 if p1 is None else p1 for q in targets}
    assert gate.bitflip_map == (p0map, p1map)


def test_measurement_einsum_string():
    func = gates.M.einsum_string
    estr = func({0, 2, 4}, 5)
    assert estr == "abcdeagcie->bdgi"
    estr = func({0, 2, 4}, 5, measuring=True)
    assert estr == "abcdeabcde->bd"
    estr = func({0, 1, 3, 5, 6}, 10, measuring=False)
    assert estr == "abcdefghijabmdofgrst->cehijmorst"
    estr = func({0, 1, 3, 5, 6}, 10, measuring=True)
    assert estr == "abcdefghijabcdefghij->cehij"


def test_measurement_add():
    gate = gates.M(0, 2)
    assert gate.target_qubits == (0, 2)
    assert gate.bitflip_map == 2 * ({0: 0, 2: 0},)
    gate.add(gates.M(1, 3, p0=0.3, p1=0.0))
    assert gate.target_qubits == (0, 2, 1, 3)
    assert gate.bitflip_map == ({0: 0, 1: 0.3, 2: 0, 3: 0.3},
                                {0: 0, 1: 0, 2: 0, 3: 0})


def test_measurement_errors():
    gate = gates.M(0)
    with pytest.raises(NotImplementedError):
        gate.controlled_by(1)


@pytest.mark.parametrize("gatename,params",
                         [("RX", (0.1234,)), ("RY", (0.1234,)),
                          ("RZ", (0.1234,)), ("U1", (0.1234,)),
                          ("U2", (0.1234, 0.4321)),
                          ("U3", (0.1234, 0.4321, 0.5678))])
def test_one_qubit_rotations_init(gatename, params):
    gate = getattr(gates, gatename)(0, *params)
    assert gate.target_qubits == (0,)
    if len(params) == 1:
        params = params[0]
    assert gate.parameters == params


@pytest.mark.parametrize("gatename,params",
                         [("RX", (0.1234,)), ("RY", (0.1234,)),
                          ("RZ", (0.1234,)), ("U1", (0.1234,)),
                          ("U2", (0.1234, 0.4321)),
                          ("U3", (0.1234, 0.4321, 0.5678))])
def test_one_qubit_rotations_controlled_by(gatename, params):
    gate = getattr(gates, gatename)(0, *params).controlled_by(1)
    assert gate.target_qubits == (0,)
    assert gate.control_qubits == (1,)
    assert isinstance(gate, getattr(gates, f"C{gatename}"))
    gate = getattr(gates, gatename)(1, *params).controlled_by(0, 3)
    assert gate.target_qubits == (1,)
    assert gate.control_qubits == (0, 3)
    if len(params) == 1:
        params = params[0]
    assert gate.parameters == params


def test_cnot_and_cz_init():
    gate = gates.CNOT(0, 1)
    assert gate.target_qubits == (1,)
    assert gate.control_qubits == (0,)
    gate = gates.CZ(3, 2)
    assert gate.target_qubits == (2,)
    assert gate.control_qubits == (3,)

# :meth:`qibo.abstractions.gates.CNOT.decompose` is tested in
# ``test_x_decompose_with_cirq`` above

@pytest.mark.parametrize("gatename,params",
                         [("CRX", (0.1234,)), ("CRY", (0.1234,)),
                          ("CRZ", (0.1234,)), ("CU1", (0.1234,)),
                          ("CU2", (0.1234, 0.4321)),
                          ("CU3", (0.1234, 0.4321, 0.5678))])
def test_two_qubit_controlled_rotations_init(gatename, params):
    gate = getattr(gates, gatename)(0, 2, *params)
    assert gate.target_qubits == (2,)
    assert gate.control_qubits == (0,)


def test_swap_init():
    gate = gates.SWAP(4, 3)
    assert gate.target_qubits == (4, 3)


def test_fsim_init():
    import numpy as np
    gate = gates.fSim(0, 1, 0.1234, 0.4321)
    assert gate.target_qubits == (0, 1)
    matrix = np.random.random((2, 2))
    gate = gates.GeneralizedfSim(0, 1, matrix, 0.4321)
    assert gate.target_qubits == (0, 1)
    assert gate.parameters == (matrix, 0.4321)
    matrix = np.random.random((3, 3))
    with pytest.raises(ValueError):
        gate = gates.GeneralizedfSim(0, 1, matrix, 0.4321)


def test_toffoli_init():
    gate = gates.TOFFOLI(0, 2, 1)
    assert gate.target_qubits == (1,)
    assert gate.control_qubits == (0, 2)

# :meth:`qibo.abstractions.gates.TOFFOLI.decompose` and
# :meth:`qibo.abstractions.gates.TOFFOLI.congruent`
# are tested in `test_x_decompose_with_cirq`

@pytest.mark.parametrize("targets", [(0,), (2, 0), (1, 3, 2)])
def test_unitary_init(targets):
    import numpy as np
    matrix = np.random.random(2 * (2 ** len(targets),))
    gate = gates.Unitary(matrix, *targets)
    assert gate.target_qubits == targets
    assert gate.nparams == 4 ** len(targets)


@pytest.mark.parametrize("targets", [range(5), range(6)])
def test_variational_layer_init(targets):
    import numpy as np
    targets = tuple(targets)
    pairs = [(i, i + 1) for i in range(0, len(targets) - 1, 2)]
    params = np.random.random(len(targets))
    gate = gates.VariationalLayer(targets, pairs, gates.RY, gates.CZ, params)
    assert gate.target_qubits == targets
    assert gate.params == {q: p for q, p in zip(targets, params)}
    assert gate.params2 == {}

    pairs = [(0, 1)]
    with pytest.raises(ValueError):
        gate = gates.VariationalLayer(targets, pairs, gates.RY, gates.CZ, params)


def test_variational_layer_errors():
    import numpy as np
    pairs = [(i, i + 1) for i in range(0, 5, 2)]
    with pytest.raises(ValueError):
        gate = gates.VariationalLayer(range(6), pairs, gates.RY, gates.CZ,
                                      np.zeros(6), np.zeros(7))
    with pytest.raises(ValueError):
        gate = gates.VariationalLayer(range(10), pairs,
                                      gates.RY, gates.CZ,
                                      np.zeros(10), np.zeros(10))
    gate = gates.VariationalLayer(range(6), pairs, gates.RY, gates.CZ,
                                  np.zeros(6), np.zeros(6))
    np.testing.assert_allclose(gate.parameters, np.zeros(12))


def test_flatten():
    gate = gates.Flatten([1, 2, 3, 4])
    assert gate.coefficients == [1, 2, 3, 4]
    with pytest.raises(NotImplementedError):
        gate._on_qubits(0, 2)


def test_kraus_channel_init():
    import numpy as np
    ops = [((0,), np.random.random((2, 2))),
           ((0, 1), np.random.random((4, 4))),
           ((0, 2), np.random.random((4, 4))),
           ((3,), np.random.random((2, 2)))]
    gate = gates.KrausChannel(ops)
    gate.target_qubits == (0, 1, 2, 3)
    for g in gate.gates:
        assert isinstance(g, gates.Unitary)

    ops.append(((4,), np.random.random((4, 4))))
    with pytest.raises(ValueError):
        gate = gates.KrausChannel(ops)


def test_unitary_channel_init():
    import numpy as np
    ops = [((0,), np.random.random((2, 2))),
           ((0, 1), np.random.random((4, 4))),
           ((0, 2), np.random.random((4, 4))),
           ((3,), np.random.random((2, 2)))]
    gate = gates.UnitaryChannel(4 * [0.1], ops)
    gate.target_qubits == (0, 1, 2, 3)
    for g in gate.gates:
        assert isinstance(g, gates.Unitary)

    with pytest.raises(ValueError):
        gate = gates.UnitaryChannel(2 * [0.1], ops)
    with pytest.raises(ValueError):
        gate = gates.UnitaryChannel(4 * [-0.1], ops)
    with pytest.raises(ValueError):
        gate = gates.UnitaryChannel(4 * [0.5], ops)


def test_pauli_noise_channel_init():
    gate = gates.PauliNoiseChannel(0, 0.1, 0.2, 0.3)
    assert gate.target_qubits == (0,)
    assert isinstance(gate.gates[0], gates.X)
    assert isinstance(gate.gates[1], gates.Y)
    assert isinstance(gate.gates[2], gates.Z)


def test_reset_channel_init():
    gate = gates.ResetChannel(0, 0.1, 0.2)
    assert gate.target_qubits == (0,)
    assert isinstance(gate.gates[0], gates.M)
    assert isinstance(gate.gates[1], gates.X)


def test_qubit_getter_and_setter():
    from qibo.abstractions import abstract_gates
    gate = abstract_gates.Gate()
    gate.target_qubits = (0, 3)
    gate.control_qubits = (1, 4, 2)
    assert gate.qubits == (1, 2, 4, 0, 3)

    gate = abstract_gates.Gate()
    with pytest.raises(ValueError):
        gate.target_qubits = (1, 1)
    gate = abstract_gates.Gate()
    with pytest.raises(ValueError):
        gate.control_qubits = (1, 1)
    gate = abstract_gates.Gate()
    gate.target_qubits = (0, 1)
    with pytest.raises(ValueError):
        gate.control_qubits = (1,)


def test_nqubits_getter_and_setter():
    from qibo.abstractions import abstract_gates
    gate = abstract_gates.Gate()
    gate.target_qubits = (0, 1)
    gate.control_qubits = (2,)
    gate.nqubits = 10
    assert gate.nqubits == 10
    assert gate.nstates == 1024


def test_nqubits_getter_and_setter_errors():
    from qibo.abstractions import abstract_gates
    gate = abstract_gates.Gate()
    with pytest.raises(ValueError):
        nqubits = gate.nqubits
    with pytest.raises(ValueError):
        nstates = gate.nstates
    gate.nqubits = 4
    with pytest.raises(ValueError):
        gate.nqubits = 5


def test_density_matrix_getter_and_setter():
    from qibo.abstractions import abstract_gates
    gate = abstract_gates.Gate()
    gate.target_qubits = (0, 1)
    gate.control_qubits = (2,)
    assert gate._active_call == "_state_vector_call"
    gate.density_matrix = True
    assert gate._active_call == "_density_matrix_call"

    gate.nqubits = 4
    gate.is_prepared = True
    with pytest.raises(RuntimeError):
        gate.density_matrix = False

def test_gates_commute():
    assert gates.H(0).commutes(gates.X(1))
    assert gates.H(0).commutes(gates.H(1))
    assert gates.H(0).commutes(gates.H(0))
    assert not gates.H(0).commutes(gates.Y(0))
    assert not gates.CNOT(0, 1).commutes(gates.SWAP(1, 2))
    assert not gates.CNOT(0, 1).commutes(gates.H(1))
    assert not gates.CNOT(0, 1).commutes(gates.Y(0).controlled_by(2))
    assert not gates.CNOT(2, 3).commutes(gates.CNOT(3, 0))
    assert gates.CNOT(0, 1).commutes(gates.Y(2).controlled_by(0))


def test_on_qubits():
    gate = gates.CNOT(0, 1)._on_qubits(2, 3)
    assert gate.target_qubits == (3,)
    assert gate.control_qubits == (2,)
    assert isinstance(gate, gates.CNOT)


def test_controlled_by():
    gate = gates.RX(0, 0.1234).controlled_by(1, 2, 3)
    assert gate.target_qubits == (0,)
    assert gate.control_qubits == (1, 2, 3)
    assert gate.is_controlled_by
    assert isinstance(gate, gates.RX)

    with pytest.raises(RuntimeError):
        gate = gates.CNOT(0, 1).controlled_by(2)

    gate = gates.RY(0, 0.1234)
    gate.nqubits = 5
    with pytest.raises(RuntimeError):
        gate = gate.controlled_by(2)


def test_on_qubits_controlled_by():
    gate = gates.H(0).controlled_by(1, 2)
    gate = gate._on_qubits(5, 4, 6)
    assert gate.target_qubits == (5,)
    assert gate.control_qubits == (4, 6)
    assert isinstance(gate, gates.H)
    assert gate.is_controlled_by


def test_decompose():
    decomp_gates = gates.H(0).decompose(1)
    assert len(decomp_gates) == 1
    assert isinstance(decomp_gates[0], gates.H)


def test_special_gate():
    from qibo.abstractions import abstract_gates
    gate = abstract_gates.SpecialGate()
    assert not gate.commutes(gates.H(0))
    with pytest.raises(NotImplementedError):
        gate._on_qubits(1)


def test_fused_gate():
    gate = gates.FusedGate(0, 1)
    gate.add(gates.H(0))
    gate.add(gates.CNOT(0, 1))
    with pytest.raises(ValueError):
        gate.add(gates.CZ(1, 2))



"""Test methods defined in `qibo/models/circuit.py`."""
import sys
import numpy as np
import pytest
from qibo import gates, models, K
from qibo.tests.utils import random_state


def test_circuit_constructor():
    from qibo.core.circuit import Circuit, DensityMatrixCircuit
    from qibo.core.distcircuit import DistributedCircuit
    c = models.Circuit(5)
    assert isinstance(c, Circuit)
    c = models.Circuit(5, density_matrix=True)
    assert isinstance(c, DensityMatrixCircuit)
    if not K.supports_multigpu:  # pragma: no cover
        with pytest.raises(NotImplementedError):
            c = models.Circuit(5, accelerators={"/GPU:0": 2})
    else:
        c = models.Circuit(5, accelerators={"/GPU:0": 2})
        assert isinstance(c, DistributedCircuit)
    with pytest.raises(NotImplementedError):
        c = models.Circuit(5, accelerators={"/GPU:0": 2}, density_matrix=True)


def qft_matrix(dimension: int, inverse: bool = False) -> np.ndarray:
    """Creates exact QFT matrix.

    Args:
        dimension: Dimension d of the matrix. The matrix will be d x d.
        inverse: Whether to construct matrix for the inverse QFT.

    Return:
        QFT transformation matrix as a numpy array with shape (d, d).
    """
    exponent = np.outer(np.arange(dimension), np.arange(dimension))
    sign = 1 - 2 * int(inverse)
    return np.exp(sign * 2 * np.pi * 1j * exponent / dimension)


def exact_qft(x: np.ndarray, inverse: bool = False) -> np.ndarray:
    """Performs exact QFT to a given state vector."""
    dim = len(x)
    return qft_matrix(dim, inverse).dot(x) / np.sqrt(dim)


@pytest.mark.parametrize("nqubits", [4, 10, 100])
def test_qft_circuit_size(backend, nqubits):
    c = models.QFT(nqubits)
    assert c.nqubits == nqubits
    assert c.depth == 2 * nqubits
    assert c.ngates == nqubits ** 2 // 2 + nqubits


@pytest.mark.parametrize("nqubits", [5, 6, 12])
@pytest.mark.parametrize("random", [False, True])
def test_qft_execution(backend, accelerators, nqubits, random):
    c = models.QFT(nqubits)
    if random:
        initial_state = random_state(nqubits)
        final_state = c(K.cast(np.copy(initial_state)))
    else:
        initial_state = c.get_initial_state()
        final_state = c()
    target_state = exact_qft(K.to_numpy(initial_state))
    K.assert_allclose(final_state, target_state)


def test_qft_errors():
    """Check that ``_DistributedQFT`` raises error if not sufficient qubits."""
    from qibo.models.circuit import _DistributedQFT
    with pytest.raises(NotImplementedError):
        c = _DistributedQFT(2, accelerators={"/GPU:0": 4})
    with pytest.raises(NotImplementedError):
        c = models.QFT(10, with_swaps=False, accelerators={"/GPU:0": 2})



import sympy
from qibo import gates, matrices, K
from qibo.config import raise_error


class Symbol(sympy.Symbol):
    """Qibo specialization for ``sympy`` symbols.

    These symbols can be used to create :class:`qibo.core.hamiltonians.SymbolicHamiltonian`.
    See :ref:`How to define custom Hamiltonians using symbols? <symbolicham-example>`
    for more details.

    Example:
        .. testcode::

            from qibo import hamiltonians
            from qibo.symbols import X, Y, Z
            # construct a XYZ Hamiltonian on two qubits using Qibo symbols
            form = X(0) * X(1) + Y(0) * Y(1) + Z(0) * Z(1)
            ham = hamiltonians.SymbolicHamiltonian(form)

    Args:
        q (int): Target qubit id.
        matrix (np.ndarray): 2x2 matrix represented by this symbol.
        name (str): Name of the symbol which defines how it is represented in
            symbolic expressions.
        commutative (bool): If ``True`` the constructed symbols commute with
            each other. Default is ``False``.
            This argument should be used with caution because quantum operators
            are not commutative objects and therefore switching this to ``True``
            may lead to wrong results. It is useful for improving performance
            in symbolic calculations in cases where the user is sure that
            the operators participating in the Hamiltonian form are commuting
            (for example when the Hamiltonian consists of Z terms only).
    """

    def __new__(cls, q, matrix=None, name="Symbol", commutative=False):
        name = "{}{}".format(name, q)
        return super().__new__(cls=cls, name=name, commutative=commutative)

    def __init__(self, q, matrix=None, name="Symbol", commutative=False):
        self.target_qubit = q
        self._gate = None
        if not (matrix is None or isinstance(matrix, K.qnp.numeric_types) or
                isinstance(matrix, K.qnp.tensor_types)):
            raise_error(TypeError, "Invalid type {} of symbol matrix."
                                   "".format(type(matrix)))
        self.matrix = matrix

    @property
    def gate(self):
        """Qibo gate that implements the action of the symbol on states."""
        if self._gate is None:
            self._gate = self.calculate_gate()
        return self._gate

    def calculate_gate(self):
        return gates.Unitary(self.matrix, self.target_qubit)

    def full_matrix(self, nqubits):
        """Calculates the full dense matrix corresponding to the symbol as part of a bigger system.

        Args:
            nqubits (int): Total number of qubits in the system.

        Returns:
            Matrix of dimension (2^nqubits, 2^nqubits) composed of the Kronecker
            product between identities and the symbol's single-qubit matrix.
        """
        from qibo.hamiltonians import multikron
        matrix_list = self.target_qubit * [matrices.I]
        matrix_list.append(self.matrix)
        n = nqubits - self.target_qubit - 1
        matrix_list.extend(matrices.I for _ in range(n))
        return multikron(matrix_list)


class PauliSymbol(Symbol):

    def __new__(cls, q, commutative=False):
        matrix = getattr(matrices, cls.__name__)
        return super().__new__(cls, q, matrix, cls.__name__, commutative)

    def __init__(self, q, commutative=False):
        name = self.__class__.__name__
        matrix = getattr(matrices, name)
        super().__init__(q, matrix, name, commutative)

    def calculate_gate(self):
        name = self.__class__.__name__
        return getattr(gates, name)(self.target_qubit)


class X(PauliSymbol):
    """Qibo symbol for the Pauli-X operator.

    Args:
        q (int): Target qubit id.
    """
    pass


class Y(PauliSymbol):
    """Qibo symbol for the Pauli-X operator.

    Args:
        q (int): Target qubit id.
    """
    pass


class Z(PauliSymbol):
    """Qibo symbol for the Pauli-X operator.

    Args:
        q (int): Target qubit id.
    """
    pass



import math
from abc import ABC, abstractmethod
from qibo import K
from qibo.abstractions import callbacks as abstract_callbacks
from qibo.abstractions.states import AbstractState
from qibo.config import EIGVAL_CUTOFF, raise_error, log


class BackendCallback(abstract_callbacks.Callback, ABC):

    def __getitem__(self, k):
        if isinstance(k, int):
            if k >= len(self._results):
                raise_error(IndexError, "Attempting to access callbacks {} run but "
                                        "the callback has been used in {} executions."
                                        "".format(k, len(self._results)))
            return self._results[k]
        if isinstance(k, slice) or isinstance(k, list) or isinstance(k, tuple):
            from qibo import K
            return K.qnp.stack(self._results[k])
        raise_error(IndexError, "Unrecognized type for index {}.".format(k))

    @abstractmethod
    def _state_vector_call(self, state): # pragma: no cover
        raise_error(NotImplementedError)

    @abstractmethod
    def _density_matrix_call(self, state): # pragma: no cover
        raise_error(NotImplementedError)

    def __call__(self, state):
        if isinstance(state, AbstractState):
            state = state.tensor
        return getattr(self, self._active_call)(state)


class EntanglementEntropy(BackendCallback, abstract_callbacks.EntanglementEntropy):

    def __init__(self, partition=None, compute_spectrum=False):
        abstract_callbacks.EntanglementEntropy.__init__(
            self, partition, compute_spectrum)
        self.partial_trace = None

    @abstract_callbacks.Callback.density_matrix.setter
    def density_matrix(self, x):
        abstract_callbacks.Callback.density_matrix.fset(self, x) # pylint: disable=no-member
        if self.partial_trace is not None:
            self.partial_trace.density_matrix = x

    @abstract_callbacks.Callback.nqubits.setter
    def nqubits(self, n: int):
        from qibo import gates
        if self._nqubits is not None and self._nqubits != n:
            raise_error(RuntimeError,
                        f"Changing EntanglementEntropy nqubits from {self._nqubits} to {n}.")
        self._nqubits = n
        if self.partition is None:
            self.partition = list(range(n // 2 + n % 2))
        if len(self.partition) <= self.nqubits // 2:
            self.partition = [i for i in range(self.nqubits)
                              if i not in set(self.partition)]
        self.partial_trace = gates.PartialTrace(*self.partition)
        self.partial_trace.nqubits = n
        self.partial_trace.density_matrix = self.density_matrix

    def entropy(self, rho):
        """Calculates entropy of a density matrix via exact diagonalization."""
        # Diagonalize
        eigvals = K.real(K.eigvalsh(rho))
        # Treating zero and negative eigenvalues
        drop_condition = eigvals > EIGVAL_CUTOFF
        masked_eigvals = K.gather(eigvals, condition=drop_condition)
        spectrum = -1 * K.log(masked_eigvals)
        if self.compute_spectrum:
            self.spectrum.append(spectrum)
        entropy = K.sum(masked_eigvals * spectrum)
        return entropy / math.log(2.0)

    def _set_nqubits(self, state):
        if not isinstance(state, K.tensor_types):
            raise_error(TypeError, "State of unknown type {} was given in callback "
                                   "calculation.".format(type(state)))
        self.nqubits = int(math.log2(tuple(state.shape)[0]))

    def _state_vector_call(self, state):
        self._set_nqubits(state)
        rho = self.partial_trace.state_vector_partial_trace(state)
        return self.entropy(rho)

    def _density_matrix_call(self, state):
        self._set_nqubits(state)
        rho = self.partial_trace.density_matrix_partial_trace(state)
        return self.entropy(rho)


class Norm(BackendCallback, abstract_callbacks.Norm):

    def _state_vector_call(self, state):
        return K.sqrt(K.sum(K.square(K.abs(state))))

    def _density_matrix_call(self, state):
        return K.trace(state)


class Overlap(BackendCallback, abstract_callbacks.Overlap):

    def __init__(self, state):
        super().__init__()
        self.statec = K.conj(K.cast(state, dtype='DTYPECPX'))

    def _state_vector_call(self, state):
        return K.abs(K.sum(self.statec * state))

    def _density_matrix_call(self, state):
        raise_error(NotImplementedError, "Overlap callback is not implemented "
                                          "for density matrices.")


class Energy(BackendCallback, abstract_callbacks.Energy):

    def _state_vector_call(self, state):
        return self.hamiltonian.expectation(state)

    def _density_matrix_call(self, state):
        return K.trace(K.matmul(self.hamiltonian.matrix, state))


class Gap(BackendCallback, abstract_callbacks.Gap):

    def __init__(self, mode="gap", check_degenerate=True):
        abstract_callbacks.Gap.__init__(self, mode, check_degenerate)
        self._evolution = None

    @property
    def evolution(self):
        """:class:`qibo.evolution.AdiabaticEvolution` model used by the callback."""
        return self._evolution

    @evolution.setter
    def evolution(self, ev: "models.AdiabaticEvolution"):
        """Sets the :class:`qibo.evolution.AdiabaticEvolution` model."""
        from qibo.models import AdiabaticEvolution
        if not isinstance(ev, AdiabaticEvolution):
            t = type(ev)
            raise_error(TypeError, "Cannot add gap callback to {}.".format(t))
        self._evolution = ev

    def _state_vector_call(self, state):
        if self.evolution is None:
            raise_error(ValueError, "Gap callback can only be used in "
                                    "adiabatic evolution models.")
        hamiltonian = self.evolution.solver.current_hamiltonian
        # Call the eigenvectors so that they are cached for the ``exp`` call
        hamiltonian.eigenvectors()
        eigvals = hamiltonian.eigenvalues()
        if isinstance(self.mode, int):
            return K.real(eigvals[self.mode])

        # case: self.mode == "gap"
        excited = 1
        gap = K.real(eigvals[excited] - eigvals[0])
        if not self.check_degenerate:
            return gap

        while K.less(gap, EIGVAL_CUTOFF):
            gap = K.real(eigvals[excited] - eigvals[0])
            excited += 1
        if excited > 1:
            log.warning("The Hamiltonian is degenerate. Using eigenvalue {} "
                        "to calculate gap.".format(excited))
        return gap

    def _density_matrix_call(self, state):
        raise_error(NotImplementedError, "Gap callback is not implemented for "
                                         "density matrices.")



"""Test functions defined in `qibo/core/distcircuit.py`."""
import pytest
import numpy as np
from qibo import gates
from qibo.core.distcircuit import DistributedCircuit
from qibo.core.distutils import DistributedQubits
from qibo.tests.test_core_distcircuit import check_device_queues


def test_transform_queue_simple(backend):
    devices = {"/GPU:0": 1, "/GPU:1": 1}
    c = DistributedCircuit(4, devices)
    c.add((gates.H(i) for i in range(4)))
    c.queues.qubits = DistributedQubits([0], c.nqubits)
    tqueue = c.queues.transform(c.queue)
    assert len(tqueue) == 6
    for i in range(3):
        assert isinstance(tqueue[i], gates.H)
        assert tqueue[i].target_qubits == (i + 1,)
    assert isinstance(tqueue[3], gates.SWAP)
    assert tqueue[3].target_qubits == (0, 1)
    assert isinstance(tqueue[4], gates.H)
    assert tqueue[4].target_qubits == (1,)
    assert isinstance(tqueue[5], gates.SWAP)
    assert tqueue[5].target_qubits == (0, 1)


def test_transform_queue_more_gates(backend):
    devices = {"/GPU:0": 2, "/GPU:1": 2}
    c = DistributedCircuit(4, devices)
    c.add(gates.H(0))
    c.add(gates.H(1))
    c.add(gates.CNOT(2, 3))
    c.add(gates.CZ(0, 1))
    c.add(gates.CNOT(3, 0))
    c.add(gates.CNOT(1, 2))
    c.queues.qubits = DistributedQubits([2, 3], c.nqubits)
    tqueue = c.queues.transform(c.queue)

    assert len(tqueue) == 10
    assert isinstance(tqueue[0], gates.H)
    assert tqueue[0].target_qubits == (0,)
    assert isinstance(tqueue[1], gates.H)
    assert tqueue[1].target_qubits == (1,)
    assert isinstance(tqueue[2], gates.CZ)
    assert tqueue[2].target_qubits == (1,)
    assert isinstance(tqueue[3], gates.SWAP)
    assert set(tqueue[3].target_qubits) == {1, 3}
    assert isinstance(tqueue[4], gates.CNOT)
    assert tqueue[4].target_qubits == (1,)
    assert isinstance(tqueue[5], gates.CNOT)
    assert tqueue[5].target_qubits == (0,)
    assert isinstance(tqueue[6], gates.SWAP)
    assert set(tqueue[6].target_qubits) == {0, 2}
    assert isinstance(tqueue[7], gates.CNOT)
    assert tqueue[7].target_qubits == (0,)
    assert isinstance(tqueue[8], gates.SWAP)
    assert set(tqueue[8].target_qubits) == {0, 2}
    assert isinstance(tqueue[9], gates.SWAP)
    assert set(tqueue[9].target_qubits) == {1, 3}


def test_create_queue_with_global_swap(backend):
    devices = {"/GPU:0": 2, "/GPU:1": 2}
    c = DistributedCircuit(6, devices)
    c.add([gates.H(0), gates.H(2), gates.H(3)])
    c.add(gates.SWAP(3, 4))
    c.add([gates.X(1), gates.X(2)])
    c.queues.qubits = DistributedQubits([4, 5], c.nqubits)
    c.queues.create(c.queues.transform(c.queue))

    check_device_queues(c.queues)
    assert len(c.queues.special_queue) == 1
    assert len(c.queues.queues) == 3
    assert len(c.queues.queues[0]) == 4
    assert len(c.queues.queues[1]) == 0
    assert len(c.queues.queues[2]) == 4
    for device_group in c.queues.queues[0]:
        assert len(device_group) == 3
    for device_group in c.queues.queues[2]:
        assert len(device_group) == 2


def test_create_queue_errors(backend):
    c = DistributedCircuit(4, {"/GPU:0": 2})
    c.add(gates.H(0))
    c.add(gates.H(1))
    c.queues.qubits = DistributedQubits([0], c.nqubits)
    with pytest.raises(ValueError):
        c.queues.create(c.queue)

    c = DistributedCircuit(4, {"/GPU:0": 4})
    c.add(gates.SWAP(0, 1))
    c.queues.qubits = DistributedQubits([0, 1], c.nqubits)
    with pytest.raises(ValueError):
        c.queues.create(c.queue)



"""Test all methods defined in `qibo/abstractions/callbacks.py`."""
import pytest
from qibo.abstractions import callbacks


def test_abstract_callback_properties():
    callback = callbacks.Callback()
    callback.nqubits = 5
    callback.append(1)
    callback.extend([2, 3])
    assert callback.nqubits == 5
    assert callback.results == [1, 2, 3]
    assert not callback.density_matrix
    assert callback._active_call == "_state_vector_call"
    callback.density_matrix = True
    assert callback.density_matrix
    assert callback._active_call == "_density_matrix_call"


def test_creating_abstract_callbacks():
    callback = callbacks.EntanglementEntropy()
    callback = callbacks.EntanglementEntropy([1, 2], compute_spectrum=True)
    callback = callbacks.Norm()
    callback = callbacks.Overlap()
    callback = callbacks.Energy("test")
    callback = callbacks.Gap()
    callback = callbacks.Gap(2)
    with pytest.raises(ValueError):
        callback = callbacks.Gap("test")
    with pytest.raises(TypeError):
        callback = callbacks.Gap(1.0)



# -*- coding: utf-8 -*-
import collections
import sympy
from abc import ABC, abstractmethod
from qibo.config import raise_error
from collections.abc import Iterable
from typing import List, Sequence, Tuple


class Gate:
    """The base class for gate implementation.

    All base gates should inherit this class.
    """

    from qibo.abstractions import gates as module

    def __init__(self):
        """
        Attributes:
            name (str): Name of the gate.
            is_controlled_by (bool): ``True`` if the gate was created using the
                :meth:`qibo.abstractions.abstract_gates.Gate.controlled_by` method,
                otherwise ``False``.
            init_args (list): Arguments used to initialize the gate.
            init_kwargs (dict): Arguments used to initialize the gate.
            target_qubits (tuple): Tuple with ids of target qubits.
            control_qubits (tuple): Tuple with ids of control qubits sorted in
                increasing order.
            nqubits (int): Number of qubits that this gate acts on.
            nstates (int): Size of state vectors that this gate acts on.
            density_matrix (bool): Controls if the gate acts on state vectors or
                density matrices.
        """
        from qibo import config
        self.name = None
        self.is_controlled_by = False
        # args for creating gate
        self.init_args = []
        self.init_kwargs = {}

        self._target_qubits = tuple()
        self._control_qubits = set()

        self._nqubits = None
        self._nstates = None
        config.ALLOW_SWITCHERS = False

        self.is_prepared = False
        self.well_defined = True
        # Keeps track of whether parametrized gates are well-defined
        # (parameter value is known during circuit creation) or if they are
        # measurement dependent so the parameter value is determined during
        # execution

        # Using density matrices or state vectors
        self._density_matrix = False
        self._active_call = "_state_vector_call"

    @property
    def target_qubits(self) -> Tuple[int]:
        """Tuple with ids of target qubits."""
        return self._target_qubits

    @property
    def control_qubits(self) -> Tuple[int]:
        """Tuple with ids of control qubits sorted in increasing order."""
        return tuple(sorted(self._control_qubits))

    @property
    def qubits(self) -> Tuple[int]:
        """Tuple with ids of all qubits (control and target) that the gate acts."""
        return self.control_qubits + self.target_qubits

    def _set_target_qubits(self, qubits: Sequence[int]):
        """Helper method for setting target qubits."""
        self._target_qubits = tuple(qubits)
        if len(self._target_qubits) != len(set(qubits)):
            repeated = self._find_repeated(qubits)
            raise_error(ValueError, "Target qubit {} was given twice for gate {}."
                                    "".format(repeated, self.name))

    def _set_control_qubits(self, qubits: Sequence[int]):
        """Helper method for setting control qubits."""
        self._control_qubits = set(qubits)
        if len(self._control_qubits) != len(qubits):
            repeated = self._find_repeated(qubits)
            raise_error(ValueError, "Control qubit {} was given twice for gate {}."
                                    "".format(repeated, self.name))

    @target_qubits.setter
    def target_qubits(self, qubits: Sequence[int]):
        """Sets target qubits tuple."""
        self._set_target_qubits(qubits)
        self._check_control_target_overlap()

    @control_qubits.setter
    def control_qubits(self, qubits: Sequence[int]):
        """Sets control qubits set."""
        self._set_control_qubits(qubits)
        self._check_control_target_overlap()

    def _set_targets_and_controls(self, target_qubits: Sequence[int],
                                 control_qubits: Sequence[int]):
        """Sets target and control qubits simultaneously.

        This is used for the reduced qubit updates in the distributed circuits
        because using the individual setters may raise errors due to temporary
        overlap of control and target qubits.
        """
        self._set_target_qubits(target_qubits)
        self._set_control_qubits(control_qubits)
        self._check_control_target_overlap()

    @staticmethod
    def _find_repeated(qubits: Sequence[int]) -> int:
        """Finds the first qubit id that is repeated in a sequence of qubit ids."""
        temp_set = set()
        for qubit in qubits:
            if qubit in temp_set:
                return qubit
            temp_set.add(qubit)

    def _check_control_target_overlap(self):
        """Checks that there are no qubits that are both target and controls."""
        common = set(self._target_qubits) & self._control_qubits
        if common:
            raise_error(ValueError, "{} qubits are both targets and controls for "
                                    "gate {}.".format(common, self.name))

    @property
    def nqubits(self) -> int:
        """Number of qubits that this gate acts on."""
        if self._nqubits is None:
            raise_error(ValueError, "Accessing number of qubits for gate {} but "
                                    "this is not yet set.".format(self))
        return self._nqubits

    @property
    def nstates(self) -> int:
        """Size of the state vectors that this gate acts on."""
        if self._nstates is None:
            raise_error(ValueError, "Accessing number of qubits for gate {} but "
                                    "this is not yet set.".format(self))
        return self._nstates

    @nqubits.setter
    def nqubits(self, n: int):
        """Sets the total number of qubits that this gate acts on.

        This setter is used by `circuit.add` if the gate is added in a circuit
        or during `__call__` if the gate is called directly on a state.
        The user is not supposed to set `nqubits` by hand.
        """
        if self._nqubits is not None and n != self.nqubits:
            raise_error(ValueError, "Cannot set gate number of qubits to {} "
                                    "because it is already set to {}."
                                    "".format(n, self.nqubits))
        self._nqubits = n
        self._nstates = 2**n

    @property
    def density_matrix(self) -> bool:
        """Controls if the gate acts on state vectors or density matrices."""
        return self._density_matrix

    @density_matrix.setter
    def density_matrix(self, x: bool):
        """Density matrix flag switcher."""
        if self.is_prepared:
            raise_error(RuntimeError,
                        "Density matrix mode cannot be switched after "
                        "preparing the gate for execution.")
        self._density_matrix = x
        if x:
            self._active_call = "_density_matrix_call"
        else:
            self._active_call = "_state_vector_call"

    def commutes(self, gate: "Gate") -> bool:
        """Checks if two gates commute.

        Args:
            gate: Gate to check if it commutes with the current gate.

        Returns:
            ``True`` if the gates commute, otherwise ``False``.
        """
        if isinstance(gate, SpecialGate):
            return False
        t1 = set(self.target_qubits)
        t2 = set(gate.target_qubits)
        a = self.__class__ == gate.__class__ and t1 == t2
        b = not (t1 & set(gate.qubits) or t2 & set(self.qubits))
        return a or b

    def _on_qubits(self, *q) -> "Gate":
        """Helper method for :meth:`qibo.abstractions.circuit.AbstractCircuit.on_qubits`.

        Creates the same gate targeting different qubits.

        Args:
            q (int): Qubit index (or indeces) that the new gate should act on.
                Note that q is interpreted as a map from the original qubit ids
                to the new ones. It is required for `len(q)` to be greater than
                the max qubit id of the original gate.

        Returns:
            A :class:`qibo.abstractions.gates.Gate` object of the original gate
            type targeting the given qubits.

        Example:

            .. testcode::

                from qibo import models, gates
                c = models.Circuit(4)
                # Add some CNOT gates
                c.add(gates.CNOT(2, 3)._on_qubits(0, 1, 2, 3)) # equivalent to gates.CNOT(2, 3)
                c.add(gates.CNOT(2, 3)._on_qubits(1, 2, 3, 0)) # equivalent to gates.CNOT(3, 0)
                c.add(gates.CNOT(2, 3)._on_qubits(2, 0, 1, 3)) # equivalent to gates.CNOT(1, 3)
                c.add(gates.CNOT(2, 3)._on_qubits(0, 3, 2, 1)) # equivalent to gates.CNOT(2, 1)
                print(c.draw())
            .. testoutput::

                q0: ───X─────
                q1: ───|─o─X─
                q2: ─o─|─|─o─
                q3: ─X─o─X───
        """
        if self.is_controlled_by:
            targets = (q[i] for i in self.target_qubits)
            controls = (q[i] for i in self.control_qubits)
            gate = self.__class__(*targets, **self.init_kwargs)
            gate = gate.controlled_by(*controls)
        else:
            qubits = (q[i] for i in self.qubits)
            gate = self.__class__(*qubits, **self.init_kwargs)
        return gate

    def _dagger(self) -> "Gate":
        """Helper method for :meth:`qibo.abstractions.gates.Gate.dagger`."""
        # By default the ``_dagger`` method creates an equivalent gate, assuming
        # that the gate is Hermitian (true for common gates like H or Paulis).
        # If the gate is not Hermitian the ``_dagger`` method should be modified.
        return self.__class__(*self.init_args, **self.init_kwargs)

    def dagger(self) -> "Gate":
        """Returns the dagger (conjugate transpose) of the gate.

        Returns:
            A :class:`qibo.abstractions.gates.Gate` object representing the dagger of
            the original gate.
        """
        new_gate = self._dagger()
        new_gate.is_controlled_by = self.is_controlled_by
        new_gate.control_qubits = self.control_qubits
        return new_gate

    def check_controls(func): # pylint: disable=E0213
        def wrapper(self, *args):
            if self.control_qubits:
                raise_error(RuntimeError, "Cannot use `controlled_by` method "
                                          "on gate {} because it is already "
                                          "controlled by {}."
                                          "".format(self, self.control_qubits))
            if self._nqubits is not None:
                raise_error(RuntimeError, "Cannot use controlled_by on a gate "
                                          "for which the number of qubits is "
                                          "set.")
            return func(self, *args) # pylint: disable=E1102
        return wrapper

    @check_controls
    def controlled_by(self, *qubits: int) -> "Gate":
        """Controls the gate on (arbitrarily many) qubits.

        Args:
            *qubits (int): Ids of the qubits that the gate will be controlled on.

        Returns:
            A :class:`qibo.abstractions.gates.Gate` object in with the corresponding
            gate being controlled in the given qubits.
        """
        if qubits:
            self.is_controlled_by = True
            self.control_qubits = qubits
        return self

    def decompose(self, *free) -> List["Gate"]:
        """Decomposes multi-control gates to gates supported by OpenQASM.

        Decompositions are based on `arXiv:9503016 <https://arxiv.org/abs/quant-ph/9503016>`_.

        Args:
            free: Ids of free qubits to use for the gate decomposition.

        Returns:
            List with gates that have the same effect as applying the original gate.
        """
        # TODO: Implement this method for all gates not supported by OpenQASM.
        # Currently this is implemented only for multi-controlled X gates.
        # If it is used on a different gate it will just return a deep copy
        # of the same gate.
        return [self.__class__(*self.init_args, **self.init_kwargs)]


class SpecialGate(Gate):
    """Abstract class for special gates.

    Current special gates are :class:`qibo.abstractions.gates.CallbackGate` and
    :class:`qibo.abstractions.gates.Flatten`.
    """

    def commutes(self, gate):
        return False

    def _on_qubits(self, *q):
        raise_error(NotImplementedError,
                    "Cannot use special gates on subroutines.")


class Channel(Gate):
    """Abstract class for channels."""

    def __init__(self):
        super().__init__()
        self.gates = tuple()
        # create inversion gates to restore the original state vector
        # because of the in-place updates used in custom operators
        self._inverse_gates = None

    @property
    def inverse_gates(self):
        if self._inverse_gates is None:
            self._inverse_gates = self.calculate_inverse_gates()
            for gate in self._inverse_gates:
                if gate is not None:
                    if self._nqubits is not None:
                        gate.nqubits = self._nqubits
                    gate.density_matrix = self.density_matrix
        return self._inverse_gates

    @abstractmethod
    def calculate_inverse_gates(self): # pragma: no cover
        raise_error(NotImplementedError)

    @Gate.nqubits.setter
    def nqubits(self, n: int):
        Gate.nqubits.fset(self, n) # pylint: disable=no-member
        for gate in self.gates:
            gate.nqubits = n
        if self._inverse_gates is not None:
            for gate in self._inverse_gates:
                if gate is not None:
                    gate.nqubits = n

    @Gate.density_matrix.setter
    def density_matrix(self, x):
        Gate.density_matrix.fset(self, x) # pylint: disable=no-member
        for gate in self.gates:
            gate.density_matrix = x
        if self._inverse_gates is not None:
            for gate in self._inverse_gates:
                if gate is not None:
                    gate.density_matrix = x

    def controlled_by(self, *q):
        """"""
        raise_error(ValueError, "Noise channel cannot be controlled on qubits.")

    def _on_qubits(self, *q): # pragma: no cover
        # future TODO
        raise_error(NotImplementedError, "`_on_qubits` method is not available "
                                         "for the `Channel` gate.")


class ParametrizedGate(Gate):
    """Base class for parametrized gates.

    Implements the basic functionality of parameter setters and getters.
    """

    def __init__(self, trainable=True):
        super(ParametrizedGate, self).__init__()
        self.parameter_names = "theta"
        self.nparams = 1
        self.trainable = trainable
        self._parameters = []
        self.symbolic_parameters = {}

    @property
    def parameters(self):
        """Returns a tuple containing the current value of gate's parameters."""
        if isinstance(self.parameter_names, str):
            return self._parameters[0]
        return tuple(self._parameters)

    @parameters.setter
    def parameters(self, x):
        """Updates the values of gate's parameters."""
        if isinstance(self.parameter_names, str):
            nparams = 1
            if not isinstance(x, collections.abc.Iterable):
                x = [x]
            else:
                # Captures the ``Unitary`` gate case where the given parameter
                # can be an array
                try:
                    if len(x) != 1:
                        x = [x]
                except TypeError: # tf.Variable case
                    s = tuple(x.shape)
                    if not s or s[0] != 1:
                        x = [x]
        else:
            nparams = len(self.parameter_names)

        if not self._parameters:
            self._parameters = nparams * [None]
        if len(x) != nparams:
            raise_error(ValueError, "Parametrized gate has {} parameters "
                                    "but {} update values were given."
                                    "".format(nparams, len(x)))
        for i, v in enumerate(x):
            if isinstance(v, sympy.Expr):
                self.well_defined = False
                self.symbolic_parameters[i] = v
            self._parameters[i] = v

        # This part uses ``BackendGate`` attributes (see below), assuming
        # that the gate was initialized using a calculation backend.
        # I could not find a cleaner way to write this so that the
        # ``circuit.set_parameters`` method works properly.
        # pylint: disable=E1101
        if isinstance(self, BaseBackendGate):
            self._reset_unitary()
            for devgate in self.device_gates:
                devgate.parameters = x

    def substitute_symbols(self):
        params = list(self._parameters)
        for i, param in self.symbolic_parameters.items():
            for symbol in param.free_symbols:
                param = symbol.evaluate(param)
            params[i] = float(param)
        self.parameters = params


class BaseBackendGate(Gate, ABC):
    """Abstract class for gate objects that can be used in calculations.
    """

    module = None

    def __init__(self):
        """
        Attributes:
            unitary: Unitary matrix representation of the gate in the computational
                basis.
            is_prepared: ``True`` if the gate is prepared for action to states.
                A gate is prepared when its matrix and/or other tensors required
                in the computation are calculated.
                See :meth:`qibo.abstractions.abstract_gates.BackendGate.prepare` for more
                details.
                Note that gate preparation is triggered automatically when a gate
                is added to a circuit or when it acts on a state.
            device: Hardware device to use in order to simulate this gate.
            density_matrix: ``True`` if the gate will act on density matrices,
                ``False`` if the gate will act on state vectors.
        """
        from qibo import get_device
        Gate.__init__(self)
        self._matrix = None
        self._cache = None
        # Cast gate matrices to the proper device
        self.device = get_device()
        # Reference to copies of this gate that are casted in devices when
        # a distributed circuit is used
        self.device_gates = set()
        self.original_gate = None

    @property
    def matrix(self):
        """Unitary matrix representing the gate in the computational basis."""
        if len(self.qubits) > 2:
            raise_error(NotImplementedError, "Cannot calculate unitary matrix for "
                                             "gates that target more than two qubits.")
        if self._matrix is None:
            self._matrix = self._construct_unitary()
        if self.is_controlled_by and tuple(self._matrix.shape) == (2, 2):
            self._matrix = self._control_unitary(self._matrix)
        return self._matrix

    def __matmul__(self, other: "Gate") -> "Gate":
        """Gate multiplication."""
        if self.qubits != other.qubits:
            raise_error(NotImplementedError, "Cannot multiply gates that target "
                                             "different qubits.")
        if self.__class__.__name__ == other.__class__.__name__:
            square_identity = {"H", "X", "Y", "Z", "CNOT", "CZ", "SWAP"}
            if self.__class__.__name__ in square_identity:
                from qibo.gates import I
                return I(*self.qubits)
        return self.module.Unitary(self.matrix @ other.matrix, *self.qubits)

    def __rmatmul__(self, other): # pragma: no cover
        # always falls back to left ``__matmul__``
        return self.__matmul__(other)

    @staticmethod
    @abstractmethod
    def _control_unitary(unitary): # pragma: no cover
        """Updates the unitary matrix of the gate if it is controlled."""
        raise_error(NotImplementedError)

    @abstractmethod
    def _construct_unitary(self): # pragma: no cover
        """Constructs the gate's unitary matrix."""
        return raise_error(NotImplementedError)

    def _reset_unitary(self):
        """Resets the gate matrices back to ``None``.

        Useful when the gate matrix need to be recalculated.
        """
        self._matrix = None

    @property
    @abstractmethod
    def cache(self): # pragma: no cover
        raise_error(NotImplementedError)

    @abstractmethod
    def _set_nqubits(self, state): # pragma: no cover
        """Sets ``gate.nqubits`` and prepares gates for application to states.

        This method is used only when gates are called directly on states
        without being a part of circuit. If a gate is added in a circuit it
        is automatically prepared and this method is not required.
        """
        raise_error(NotImplementedError)

    @abstractmethod
    def _state_vector_call(self, state): # pragma: no cover
        """Applies the gate on a state vector."""
        raise_error(NotImplementedError)

    @abstractmethod
    def _density_matrix_call(self, state): # pragma: no cover
        """Applies the gate on a density matrix."""
        raise_error(NotImplementedError)

    @abstractmethod
    def _density_matrix_half_call(self, state): # pragma: no cover
        """Half application of gate to density matrix.

        For an arbitrary unitary gate U the
        :meth:`qibo.abstractions.abstract_gates.BaseBackendGate._density_matrix_call`
        calculates

        .. math::
            U\\rho U^\\dagger

        while this method calculates only

        .. math::
            U\\rho

        This is useful for :class:`qibo.abstractions.hamiltonians.SymbolicHamiltonian`
        multiplication to density matrices.
        """
        raise_error(NotImplementedError)

    def __call__(self, state):
        """Applies the gate on a state.

        Falls back to a state vector or density matrix call according to the
        current value of the ``gate.density_matrix`` flag.
        It automatically prepares the gate if it is not already prepared.
        """
        if not self.is_prepared:
            self._set_nqubits(state)
        if not self.well_defined:
            self.substitute_symbols() # pylint: disable=E1101
            # method available only for parametrized gates
        return getattr(self, self._active_call)(state)



import os
from qibo import config
from qibo.config import raise_error, log


class Backend:

    def __init__(self):
        # load profile from default file
        from pathlib import Path
        profile_path = Path(os.environ.get(
            'QIBO_PROFILE', Path(__file__).parent / "profiles.yml"))
        try:
            with open(profile_path) as f:
                import yaml
                self.profile = yaml.safe_load(f)
        except FileNotFoundError:  # pragma: no cover
            raise_error(FileNotFoundError, f"Profile file {profile_path} not found.")

        # dictionary to cache if backends are available
        # used by ``self.check_availability``
        self._availability = {}

        # create numpy backend (is always available as numpy is a requirement)
        if self.check_availability("numpy", check_version=False):
            from qibo.backends.numpy import NumpyBackend
            self.qnp = NumpyBackend()
        else:  # pragma: no cover
            raise_error(ModuleNotFoundError, "Numpy is not installed. "
                                             "Please install it using "
                                             "`pip install numpy`.")

        # loading backend names and version
        self._backends_min_version = {backend.get("name"): backend.get("minimum_version") for backend in self.profile.get("backends")}

        # find the default backend name
        default_backend = os.environ.get('QIBO_BACKEND', self.profile.get('default'))

        # check if default backend is described in the profile file
        if default_backend != "numpy":

            if default_backend not in (backend.get("name") for backend in self.profile.get("backends")): # pragma: no cover
                raise_error(ModuleNotFoundError, f"Default backend {default_backend} not set in {profile_path}.")

            # change the default backend if it is not available
            if not self.check_availability(default_backend):  # pragma: no cover
                # set the default backend to the first available declared backend in the profile file
                # if none is available it falls back to numpy
                for backend in self.profile.get('backends'):
                    name = backend.get('name')
                    if self.check_availability(name) and not backend.get('is_hardware'):
                        # excluding hardware backends from default for development
                        # convenience until proper hardware testing is implemented
                        default_backend = name
                        break
                    # make numpy default if no other backend is available
                    default_backend = "numpy"

        self.active_backend = None
        self.constructed_backends = {"numpy": self.qnp}
        self.hardware_backends = {}
        # set default backend as active
        self.active_backend = self.construct_backend(default_backend)

        # raise performance warning if qibojit and qibotf are not available
        self.show_config()
        if str(self) == "numpy":  # pragma: no cover
            log.warning("numpy backend uses `np.einsum` and supports CPU only. "
                        "Consider installing the qibojit or qibotf backends for "
                        "increased performance and to enable GPU acceleration.")
        elif str(self) == "tensorflow":  # pragma: no cover
            # case not tested because CI has tf installed
            log.warning("qibotf library was not found. `tf.einsum` will be "
                        "used to apply gates. In order to install Qibo's "
                        "high performance custom operators for TensorFlow "
                        "please use `pip install qibotf`. Alternatively, "
                        "consider installing the qibojit backend.")

    @staticmethod
    def _get_backend_class(backend):
        """Loads class associated with the given backend.

        Helper method for ``construct_backend``.

        Args:
            backend (dict): The backend dictionary as read from ``profiles.yml``.

        Returns:
            Class that is used to initialize the given backend.
        """
        import importlib
        components = backend.get('driver').split('.')
        module, clsname = '.'.join(components[:-1]), components[-1]
        backend_module = importlib.import_module(module)
        return getattr(backend_module, clsname)

    def construct_backend(self, name):
        """Constructs and returns a backend.

        If the backend already exists in previously constructed backends then
        the existing object is returned.

        Args:
            name (str): Name of the backend to construct.
                See ``available_backends`` for the list of supported names.

        Returns:
            Backend object.
        """
        if name not in self.constructed_backends:
            backend = None
            if self.check_availability(name):
                for backend in self.profile.get('backends'):
                    if backend.get('name') == name:
                        break
            if backend is not None and backend.get('name') == name:
                backend_instance = self._get_backend_class(backend)()
                if self.active_backend is not None:
                    backend_instance.set_precision(self.active_backend.precision)
                self.constructed_backends[name] = backend_instance
                if backend.get('is_hardware', False):  # pragma: no cover
                    self.hardware_backends[name] = backend

            else:
                available = []
                for backend in self.profile.get('backends'):
                    n = backend.get('name')
                    if self.check_availability(n):
                        d = self._get_backend_class(backend).description
                        available.append(f" - {n}: {d}")
                available.append(f" - numpy: {self.qnp.description}")
                available = "\n".join(available)
                raise_error(ValueError, "Unknown backend {}. Please select one "
                                        "of the available backends:\n{}"
                                        "".format(name, available))

        return self.constructed_backends.get(name)

    def __getattr__(self, x):
        return getattr(self.active_backend, x)

    def __str__(self):
        return self.active_backend.name

    def __repr__(self):
        return str(self)

    def show_config(self):
        log.info(f"Using {self} backend on {self.active_backend.default_device}")

    def check_availability(self, module_name, check_version=True):
        """Check if module is installed.

        Args:
            module_name (str): module name.
            check_version (str): check if module has the required minimum version.
                A `ModuleNotFoundError` is raised if version is lower than minimum.

        Returns:
            ``True`` if the module is installed, ``False`` otherwise.
        """
        if module_name not in self._availability:
            from pkgutil import iter_modules
            is_available = module_name in (name for _, name, _ in iter_modules())
            if is_available and check_version:
                from importlib_metadata import version
                from packaging.version import parse
                minimum_version = self._backends_min_version.get(module_name, None)
                if minimum_version is not None and (parse(version(module_name)) < parse(minimum_version)): # pragma: no cover
                    raise_error(ModuleNotFoundError, f"Please upgrade {module_name}. "
                                                     f"Minimum supported version {minimum_version}.")
            self._availability[module_name] = is_available
        return self._availability.get(module_name)


K = Backend()
numpy_matrices = K.qnp.matrices


def set_backend(backend="qibojit"):
    """Sets backend used for mathematical operations and applying gates.

    The following backends are available:
    'qibojit': Numba/cupy backend with custom operators for applying gates,
    'qibotf': Tensorflow backend with custom operators for applying gates,
    'tensorflow': Tensorflow backend that applies gates using ``tf.einsum``,
    'numpy': Numpy backend that applies gates using ``np.einsum``.

    Args:
        backend (str): A backend from the above options.
    """
    if not config.ALLOW_SWITCHERS and backend != K.name:
        log.warning("Backend should not be changed after allocating gates.")
    K.active_backend = K.construct_backend(backend)
    K.show_config()


def get_backend():
    """Get backend used to implement gates.

    Returns:
        A string with the backend name.
    """
    return K.name


def set_precision(dtype="double"):
    """Set precision for states and gates simulation.

    Args:
        dtype (str): possible options are 'single' for single precision
            (complex64) and 'double' for double precision (complex128).
    """
    if not config.ALLOW_SWITCHERS and dtype != K.precision:
        log.warning("Precision should not be changed after allocating gates.")
    for bk in K.constructed_backends.values():
        bk.set_precision(dtype)


def get_precision():
    """Get precision for states and gates simulation.

    Returns:
        A string with the precision name ('single', 'double').
    """
    return K.precision


def set_device(name):
    """Set default execution device.

    Args:
        name (str): Device name. Should follow the pattern
            '/{device type}:{device number}' where device type is one of
            CPU or GPU.
    """
    if not config.ALLOW_SWITCHERS and name != K.default_device:
        log.warning("Device should not be changed after allocating gates.")
    K.set_device(name)
    for bk in K.constructed_backends.values():
        if bk.name != "numpy" and bk != K.active_backend:
            bk.set_device(name)


def get_device():
    return K.default_device


def set_threads(nthreads):
    """Set number of CPU threads.

    Args:
        nthreads (int): number of threads.
    """
    K.set_threads(nthreads)


def get_threads():
    """Returns number of threads."""
    return K.nthreads



from qibo import K
from qibo.config import raise_error
from qibo.core.circuit import Circuit
from qibo.models.evolution import StateEvolution


class VQE(object):
    """This class implements the variational quantum eigensolver algorithm.

    Args:
        circuit (:class:`qibo.abstractions.circuit.AbstractCircuit`): Circuit that
            implements the variaional ansatz.
        hamiltonian (:class:`qibo.hamiltonians.Hamiltonian`): Hamiltonian object.

    Example:
        .. testcode::

            import numpy as np
            from qibo import gates, models, hamiltonians
            # create circuit ansatz for two qubits
            circuit = models.Circuit(2)
            circuit.add(gates.RY(0, theta=0))
            # create XXZ Hamiltonian for two qubits
            hamiltonian = hamiltonians.XXZ(2)
            # create VQE model for the circuit and Hamiltonian
            vqe = models.VQE(circuit, hamiltonian)
            # optimize using random initial variational parameters
            initial_parameters = np.random.uniform(0, 2, 1)
            vqe.minimize(initial_parameters)
    """
    from qibo import optimizers

    def __init__(self, circuit, hamiltonian):
        """Initialize circuit ansatz and hamiltonian."""
        self.circuit = circuit
        self.hamiltonian = hamiltonian

    def minimize(self, initial_state, method='Powell', jac=None, hess=None,
                 hessp=None, bounds=None, constraints=(), tol=None, callback=None,
                 options=None, compile=False, processes=None):
        """Search for parameters which minimizes the hamiltonian expectation.

        Args:
            initial_state (array): a initial guess for the parameters of the
                variational circuit.
            method (str): the desired minimization method.
                See :meth:`qibo.optimizers.optimize` for available optimization
                methods.
            jac (dict): Method for computing the gradient vector for scipy optimizers.
            hess (dict): Method for computing the hessian matrix for scipy optimizers.
            hessp (callable): Hessian of objective function times an arbitrary
                vector for scipy optimizers.
            bounds (sequence or Bounds): Bounds on variables for scipy optimizers.
            constraints (dict): Constraints definition for scipy optimizers.
            tol (float): Tolerance of termination for scipy optimizers.
            callback (callable): Called after each iteration for scipy optimizers.
            options (dict): a dictionary with options for the different optimizers.
            compile (bool): whether the TensorFlow graph should be compiled.
            processes (int): number of processes when using the paralle BFGS method.

        Return:
            The final expectation value.
            The corresponding best parameters.
            The optimization result object. For scipy methods it returns
            the ``OptimizeResult``, for ``'cma'`` the ``CMAEvolutionStrategy.result``,
            and for ``'sgd'`` the options used during the optimization.
        """
        def _loss(params, circuit, hamiltonian):
            circuit.set_parameters(params)
            final_state = circuit()
            return hamiltonian.expectation(final_state)

        if compile:
            if K.is_custom:
                raise_error(RuntimeError, "Cannot compile VQE that uses custom operators. "
                                          "Set the compile flag to False.")
            for gate in self.circuit.queue:
                _ = gate.cache
            loss = K.compile(_loss)
        else:
            loss = _loss

        if method == "cma":
            dtype = getattr(K.np, K._dtypes.get('DTYPE'))
            loss = lambda p, c, h: dtype(_loss(p, c, h))
        elif method != "sgd":
            loss = lambda p, c, h: K.to_numpy(_loss(p, c, h))

        result, parameters, extra = self.optimizers.optimize(loss, initial_state,
                                                             args=(self.circuit, self.hamiltonian),
                                                             method=method, jac=jac, hess=hess, hessp=hessp,
                                                             bounds=bounds, constraints=constraints,
                                                             tol=tol, callback=callback, options=options,
                                                             compile=compile, processes=processes)
        self.circuit.set_parameters(parameters)
        return result, parameters, extra


class AAVQE(object):
    """This class implements the Adiabatically Assisted Variational Quantum Eigensolver
    algorithm. See https://arxiv.org/abs/1806.02287.

    Args:
        circuit (:class:`qibo.abstractions.circuit.AbstractCircuit`): variaional ansatz.
        easy_hamiltonian (:class:`qibo.hamiltonians.Hamiltonian`): initial Hamiltonian object.
        problem_hamiltonian (:class:`qibo.hamiltonians.Hamiltonian`): problem Hamiltonian object.
        s (callable): scheduling function of time that defines the adiabatic
            evolution. It must verify boundary contitions: s(0) = 0 and s(1) = 1.
        nsteps (float): number of steps of the adiabatic evolution.
        t_max (float): total time evolution.
        bounds_tolerance (float): tolerance for checking s(0) = 0 and s(T) = 1.
        time_tolerance (float): tolerance for checking if time is greater than t_max.

    Example:
    ::
        import numpy as np
        from qibo import gates, models, hamiltonians
        # create circuit ansatz for two qubits
        circuit = models.Circuit(2)
        circuit.add(gates.RY(0, theta=0))
        circuit.add(gates.RY(1, theta=0))
        # define the easy and the problem Hamiltonians.
        easy_hamiltonian=hamiltonians.X(2)
        problem_hamiltonian=hamiltonians.XXZ(2)
        # define a scheduling function with only one parameter 
        # and boundary conditions s(0) = 0, s(1) = 1
        s = lambda t: t
        # create AAVQE model 
        aavqe = models.AAVQE(circuit, easy_hamiltonian, problem_hamiltonian, 
                             s, nsteps=10, t_max=1)
        # optimize using random initial variational parameters
        np.random.seed(0)
        initial_parameters = np.random.uniform(0, 2*np.pi, 2)
        ground_energy, params = aavqe.minimize(initial_parameters)
    """

    def __init__(self, circuit, easy_hamiltonian, problem_hamiltonian, s, nsteps=10, t_max = 1, bounds_tolerance = 1e-7, time_tolerance = 1e-7):

        if nsteps <= 0: # pragma: no cover
            raise_error(ValueError, "Number of steps nsteps should be positive but is {}."
                                    "".format(nsteps))
        if t_max <= 0: # pragma: no cover
            raise_error(ValueError, "Maximum time t_max should be positive but is {}."
                                    "".format(t_max))
        if easy_hamiltonian.nqubits != problem_hamiltonian.nqubits: # pragma: no cover
            raise_error(ValueError, "The easy Hamiltonian has {} qubits while problem Hamiltonian has {}."
                                    "".format(easy_hamiltonian.nqubits, problem_hamiltonian.nqubits))

        self.ATOL = bounds_tolerance 
        self.ATOL_TIME = time_tolerance 

        self._circuit = circuit
        self._h0 = easy_hamiltonian
        self._h1 = problem_hamiltonian
        self._nsteps = nsteps
        self._t_max = t_max
        self._dt = 1./(nsteps-1)

        self._schedule = None
        nparams = s.__code__.co_argcount
        if not nparams == 1: # pragma: no cover
            raise_error(ValueError,"Scheduling function must take only one argument,"
                                   "but the function proposed takes {}.".format(nparams))
        self.set_schedule(s)
    
    def set_schedule(self, func):
        """ Set scheduling function s(t) as func."""
        #check boundary conditions
        s0 = func(0)
        if abs(s0) > self.ATOL: # pragma: no cover
            raise_error(ValueError, "s(0) should be 0 but it is {}.".format(s0))
        s1 = func(1)
        if abs(s1 - 1) > self.ATOL: # pragma: no cover
            raise_error(ValueError, "s(1) should be 1 but it is {}.".format(s1))
        self._schedule = func

    def schedule(self, t):
        """ Returns scheduling function evaluated at time t: s(t/Tmax)."""
        if self._schedule is None: # pragma: no cover
            raise_error(ValueError, "Cannot access scheduling before it is set.")
        if (t - self._t_max) > self.ATOL_TIME: # pragma: no cover
            raise_error(ValueError, "t cannot be greater than {}, but it is {}.".format(self._t_max, t))

        s = self._schedule(t / self._t_max)
        if (abs(s) - 1) > self.ATOL: # pragma: no cover
            raise_error(ValueError, "s cannot be greater than 1 but it is {}.".format(s))
        return s

    def hamiltonian(self, t):
        """Returns the adiabatic evolution Hamiltonian at a given time."""
        if (t - self._t_max) > self.ATOL: # pragma: no cover
            raise_error(ValueError, "t cannot be greater than {}, but it is {}.".format(self._t_max, t))
        # boundary conditions  s(0)=0, s(total_time)=1
        st = self.schedule(t)
        return self._h0 * (1 - st) + self._h1 * st

    def minimize(self, params, method="BFGS", jac=None, hess=None,
                 hessp=None, bounds=None, constraints=(), tol=None,
                 options=None, compile=False, processes=None):
        """
        Performs minimization to find the ground state of the problem Hamiltonian.
        Args:
            params (np.ndarray or list): initial guess for the parameters of the variational circuit.
            method (str): optimizer to employ. 
            jac (dict): Method for computing the gradient vector for scipy optimizers.
            hess (dict): Method for computing the hessian matrix for scipy optimizers.
            hessp (callable): Hessian of objective function times an arbitrary
                            vector for scipy optimizers.
            bounds (sequence or Bounds): Bounds on variables for scipy optimizers.
            constraints (dict): Constraints definition for scipy optimizers.
            tol (float): Tolerance of termination for scipy optimizers.
            options (dict): a dictionary with options for the different optimizers.
            compile (bool): whether the TensorFlow graph should be compiled.
            processes (int): number of processes when using the paralle BFGS method.
        """
        from qibo import models
        t = 0.
        while (t-self._t_max)<=self.ATOL_TIME:
            H = self.hamiltonian(t)
            vqe = models.VQE(self._circuit, H)
            best, params, _ = vqe.minimize(params, method=method, jac=jac, hess=hess, hessp=hessp, 
                                        bounds=bounds, constraints=constraints, tol=tol, 
                                        options=options, compile=compile, processes=processes)
            t += self._dt
        return best, params


class QAOA(object):
    """ Quantum Approximate Optimization Algorithm (QAOA) model.

    The QAOA is introduced in `arXiv:1411.4028 <https://arxiv.org/abs/1411.4028>`_.

    Args:
        hamiltonian (:class:`qibo.abstractions.hamiltonians.Hamiltonian`): problem Hamiltonian
            whose ground state is sought.
        mixer (:class:`qibo.abstractions.hamiltonians.Hamiltonian`): mixer Hamiltonian.
            If ``None``, :class:`qibo.hamiltonians.X` is used.
        solver (str): solver used to apply the exponential operators.
            Default solver is 'exp' (:class:`qibo.solvers.Exponential`).
        callbacks (list): List of callbacks to calculate during evolution.
        accelerators (dict): Dictionary of devices to use for distributed
            execution. See :class:`qibo.core.distcircuit.DistributedCircuit`
            for more details. This option is available only when ``hamiltonian``
            is a :class:`qibo.abstractions.hamiltonians.SymbolicHamiltonian`.

    Example:
        .. testcode::

            import numpy as np
            from qibo import models, hamiltonians
            # create XXZ Hamiltonian for four qubits
            hamiltonian = hamiltonians.XXZ(4)
            # create QAOA model for this Hamiltonian
            qaoa = models.QAOA(hamiltonian)
            # optimize using random initial variational parameters
            # and default options and initial state
            initial_parameters = 0.01 * np.random.random(4)
            best_energy, final_parameters, extra = qaoa.minimize(initial_parameters, method="BFGS")
    """
    from qibo import hamiltonians, optimizers
    from qibo.core import states

    def __init__(self, hamiltonian, mixer=None, solver="exp", callbacks=[],
                 accelerators=None):
        from qibo.abstractions.hamiltonians import AbstractHamiltonian
        # list of QAOA variational parameters (angles)
        self.params = None
        # problem hamiltonian
        if not isinstance(hamiltonian, AbstractHamiltonian):
            raise_error(TypeError, "Invalid Hamiltonian type {}."
                                   "".format(type(hamiltonian)))
        self.hamiltonian = hamiltonian
        self.nqubits = hamiltonian.nqubits
        # mixer hamiltonian (default = -sum(sigma_x))
        if mixer is None:
            trotter = isinstance(
                self.hamiltonian, self.hamiltonians.SymbolicHamiltonian)
            self.mixer = self.hamiltonians.X(self.nqubits, dense=not trotter)
        else:
            if type(mixer) != type(hamiltonian):
                  raise_error(TypeError, "Given Hamiltonian is of type {} "
                                         "while mixer is of type {}."
                                         "".format(type(hamiltonian),
                                                   type(mixer)))
            self.mixer = mixer

        # create circuits for Trotter Hamiltonians
        if (accelerators is not None and (
            not isinstance(self.hamiltonian, self.hamiltonians.SymbolicHamiltonian)
            or solver != "exp")):
            raise_error(NotImplementedError, "Distributed QAOA is implemented "
                                             "only with SymbolicHamiltonian and "
                                             "exponential solver.")
        if isinstance(self.hamiltonian, self.hamiltonians.SymbolicHamiltonian):
            self.hamiltonian.circuit(1e-2, accelerators)
            self.mixer.circuit(1e-2, accelerators)

        # evolution solvers
        from qibo import solvers
        self.ham_solver = solvers.factory[solver](1e-2, self.hamiltonian)
        self.mix_solver = solvers.factory[solver](1e-2, self.mixer)

        self.state_cls = self.states.VectorState
        self.callbacks = callbacks
        self.accelerators = accelerators
        self.normalize_state = StateEvolution._create_normalize_state(
            self, solver)
        self.calculate_callbacks = StateEvolution._create_calculate_callbacks(
            self, accelerators)

    def set_parameters(self, p):
        """Sets the variational parameters.

        Args:
            p (np.ndarray): 1D-array holding the new values for the variational
                parameters. Length should be an even number.
        """
        self.params = p

    def _apply_exp(self, state, solver, p):
        """Helper method for ``execute``."""
        solver.dt = p
        state = solver(state)
        if self.callbacks:
            state = self.normalize_state(state)
            self.calculate_callbacks(state)
        return state

    def execute(self, initial_state=None):
        """Applies the QAOA exponential operators to a state.

        Args:
            initial_state (np.ndarray): Initial state vector.

        Returns:
            State vector after applying the QAOA exponential gates.
        """
        state = self.get_initial_state(initial_state)
        self.calculate_callbacks(state)
        n = int(self.params.shape[0])
        for i in range(n // 2):
            state = self._apply_exp(state, self.ham_solver,
                                    self.params[2 * i])
            state = self._apply_exp(state, self.mix_solver,
                                    self.params[2 * i + 1])
        return self.normalize_state(state)

    def __call__(self, initial_state=None):
        """Equivalent to :meth:`qibo.models.QAOA.execute`."""
        return self.execute(initial_state)

    def get_initial_state(self, state=None):
        """"""
        if self.accelerators is not None:
            c = self.hamiltonian.circuit(self.params[0])
            if state is None:
                state = self.states.DistributedState.plus_state(c)
            return c.get_initial_state(state)

        if state is None:
            return self.state_cls.plus_state(self.nqubits).tensor
        return Circuit.get_initial_state(self, state)

    def minimize(self, initial_p, initial_state=None, method='Powell',
                 jac=None, hess=None, hessp=None, bounds=None, constraints=(),
                 tol=None, callback=None, options=None, compile=False, processes=None):
        """Optimizes the variational parameters of the QAOA.

        Args:
            initial_p (np.ndarray): initial guess for the parameters.
            initial_state (np.ndarray): initial state vector of the QAOA.
            method (str): the desired minimization method.
                See :meth:`qibo.optimizers.optimize` for available optimization
                methods.
            jac (dict): Method for computing the gradient vector for scipy optimizers.
            hess (dict): Method for computing the hessian matrix for scipy optimizers.
            hessp (callable): Hessian of objective function times an arbitrary
                vector for scipy optimizers.
            bounds (sequence or Bounds): Bounds on variables for scipy optimizers.
            constraints (dict): Constraints definition for scipy optimizers.
            tol (float): Tolerance of termination for scipy optimizers.
            callback (callable): Called after each iteration for scipy optimizers.
            options (dict): a dictionary with options for the different optimizers.
            compile (bool): whether the TensorFlow graph should be compiled.
            processes (int): number of processes when using the paralle BFGS method.

        Return:
            The final energy (expectation value of the ``hamiltonian``).
            The corresponding best parameters.
            The optimization result object. For scipy methods it
            returns the ``OptimizeResult``, for ``'cma'`` the
            ``CMAEvolutionStrategy.result``, and for ``'sgd'``
            the options used during the optimization.
        """
        if len(initial_p) % 2 != 0:
            raise_error(ValueError, "Initial guess for the parameters must "
                                    "contain an even number of values but "
                                    "contains {}.".format(len(initial_p)))

        def _loss(params, qaoa, hamiltonian):
            qaoa.set_parameters(params)
            state = qaoa(initial_state)
            return hamiltonian.expectation(state)

        if method == "sgd":
            loss = lambda p, c, h: _loss(K.cast(p), c, h)
        else:
            loss = lambda p, c, h: K.to_numpy(_loss(p, c, h))

        result, parameters, extra = self.optimizers.optimize(loss, initial_p, args=(self, self.hamiltonian),
                                                             method=method, jac=jac, hess=hess, hessp=hessp,
                                                             bounds=bounds, constraints=constraints,
                                                             tol=tol, callback=callback, options=options,
                                                             compile=compile, processes=processes)
        self.set_parameters(parameters)
        return result, parameters, extra


class FALQON(QAOA):
    """ Feedback-based ALgorithm for Quantum OptimizatioN (FALQON) model.

    The FALQON is introduced in `arXiv:2103.08619 <https://arxiv.org/abs/2103.08619>`_.
    It inherits the QAOA class.

    Args:
        hamiltonian (:class:`qibo.abstractions.hamiltonians.Hamiltonian`): problem Hamiltonian
            whose ground state is sought.
        mixer (:class:`qibo.abstractions.hamiltonians.Hamiltonian`): mixer Hamiltonian.
            If ``None``, :class:`qibo.hamiltonians.X` is used.
        solver (str): solver used to apply the exponential operators.
            Default solver is 'exp' (:class:`qibo.solvers.Exponential`).
        callbacks (list): List of callbacks to calculate during evolution.
        accelerators (dict): Dictionary of devices to use for distributed
            execution. See :class:`qibo.tensorflow.distcircuit.DistributedCircuit`
            for more details. This option is available only when ``hamiltonian``
            is a :class:`qibo.abstractions.hamiltonians.SymbolicHamiltonian`.

    Example:
        .. testcode::

            import numpy as np
            from qibo import models, hamiltonians
            # create XXZ Hamiltonian for four qubits
            hamiltonian = hamiltonians.XXZ(4)
            # create FALQON model for this Hamiltonian
            falqon = models.FALQON(hamiltonian)
            # optimize using random initial variational parameters
            # and default options and initial state
            delta_t = 0.01
            max_layers = 3
            best_energy, final_parameters, extra = falqon.minimize(delta_t, max_layers)
    """

    def __init__(self, hamiltonian, mixer=None, solver="exp", callbacks=[],
                 accelerators=None):
        super().__init__(hamiltonian, mixer, solver, callbacks, accelerators)
        self.evol_hamiltonian = 1j * (self.hamiltonian @ self.mixer - self.mixer @ self.hamiltonian)

    def minimize(self, delta_t, max_layers, initial_state=None, tol=None, callback=None):
        """Optimizes the variational parameters of the FALQON.

        Args:
            delta_t (float): initial guess for the time step. A too large delta_t will make the algorithm fail.
            max_layers (int): maximum number of layers allowed for the FALQON.
            initial_state (np.ndarray): initial state vector of the FALQON.
            tol (float): Tolerance of energy change. If not specified, no check is done.
            callback (callable): Called after each iteration for scipy optimizers.
            options (dict): a dictionary with options for the different optimizers.

        Return:
            The final energy (expectation value of the ``hamiltonian``).
            The corresponding best parameters.
            extra: variable with historical data for the energy and callbacks.
        """
        import numpy as np
        parameters = np.array([delta_t, 0])

        def _loss(params, falqon, hamiltonian):
            falqon.set_parameters(params)
            state = falqon(initial_state)
            return hamiltonian.expectation(state)

        energy = [np.inf]
        callback_result = []
        for it in range(1, max_layers + 1):
            beta = K.to_numpy(_loss(parameters, self, self.evol_hamiltonian))

            if tol is not None:
                energy.append(K.to_numpy(_loss(parameters, self, self.hamiltonian)))
                if abs(energy[-1] - energy[-2]) < tol:
                    break

            if callback is not None:
                callback_result.append(callback(parameters))

            parameters = np.concatenate([parameters, [delta_t, delta_t * beta]])

        self.set_parameters(parameters)
        final_loss = _loss(parameters, self, self.hamiltonian)
        extra = {'energies': energy, 'callbacks': callback_result}
        return final_loss, parameters, extra



# -*- coding: utf-8 -*-
# @authors: S. Carrazza and A. Garcia
import collections
from abc import ABC, abstractmethod
from qibo.abstractions import gates
from qibo import gates as gate_module
from qibo.config import raise_error
from typing import Dict, List, Optional, Tuple, Union
NoiseMapType = Union[Tuple[int, int, int],
                     Dict[int, Tuple[int, int, int]]]


class _ParametrizedGates(list):
    """Simple data structure for keeping track of parametrized gates.

    Useful for the ``circuit.set_parameters()`` method.
    Holds parametrized gates in a list and a set and also keeps track of the
    total number of parameters.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set = set()
        self.nparams = 0

    def append(self, gate):
        super().append(gate)
        self.set.add(gate)
        self.nparams += gate.nparams


class _Queue(list):
    """List that holds the queue of gates of a circuit.

    In addition to the queue, it holds a list of gate moments, where each gate
    is placed in the earliest possible position depending for the qubits it acts.
    """

    def __init__(self, nqubits):
        super(_Queue, self).__init__(self)
        self.nqubits = nqubits
        self.moments = [nqubits * [None]]
        self.moment_index = nqubits * [0]

    def append(self, gate: gates.Gate):
        super(_Queue, self).append(gate)
        # Calculate moment index for this gate
        if gate.qubits:
            idx = max(self.moment_index[q] for q in gate.qubits)
        for q in gate.qubits:
            if idx >= len(self.moments):
                # Add a moment
                self.moments.append(len(self.moments[-1]) * [None])
            self.moments[idx][q] = gate
            self.moment_index[q] = idx + 1


class AbstractCircuit(ABC):
    """Circuit object which holds a list of gates.

    This circuit is symbolic and cannot perform calculations.
    A specific backend has to be used for performing calculations.
    All backend-based circuits should inherit ``AbstractCircuit``.

    Qibo provides the following circuits:

    * A state vector simulation circuit:
      :class:`qibo.core.circuit.Circuit`.
    * A density matrix simulation circuit:
      :class:`qibo.core.circuit.DensityMatrixCircuit`.
    * A circuit that distributes state vector simulation on multiple devices:
      :class:`qibo.core.distcircuit.DistributedCircuit`.

    All circuits use core as the computation backend.

    Args:
        nqubits (int): Total number of qubits in the circuit.
    """

    def __init__(self, nqubits):
        if not isinstance(nqubits, int):
            raise_error(TypeError, "Number of qubits must be an integer but "
                                   "is {}.".format(nqubits))
        if nqubits < 1:
            raise_error(ValueError, "Number of qubits must be positive but is "
                                    "{}.".format(nqubits))
        self.nqubits = nqubits
        self.init_kwargs = {"nqubits": nqubits}
        self.queue = _Queue(nqubits)
        # Keep track of parametrized gates for the ``set_parameters`` method
        self.parametrized_gates = _ParametrizedGates()
        self.trainable_gates = _ParametrizedGates()

        self.measurement_tuples = dict()
        self.measurement_gate = None
        self.measurement_gate_result = None

        self._final_state = None
        self.density_matrix = False
        self.repeated_execution = False

        self.param_tensor_types = (None.__class__,)

    def __add__(self, circuit):
        """Add circuits.

        Args:
            circuit: Circuit to be added to the current one.

        Returns:
            The resulting circuit from the addition.
        """
        for k, kwarg1 in self.init_kwargs.items():
            kwarg2 = circuit.init_kwargs[k]
            if kwarg1 != kwarg2:
                raise_error(ValueError,
                            "Cannot add circuits with different kwargs. "
                            "{} is {} for first circuit and {} for the "
                            "second.".format(k, kwarg1, kwarg2))

        newcircuit = self.__class__(**self.init_kwargs)
        # Add gates from `self` to `newcircuit` (including measurements)
        for gate in self.queue:
            newcircuit.queue.append(gate)
            if isinstance(gate, gates.ParametrizedGate):
                newcircuit.parametrized_gates.append(gate)
                if gate.trainable:
                    newcircuit.trainable_gates.append(gate)
        newcircuit.measurement_gate = self.measurement_gate
        newcircuit.measurement_tuples = self.measurement_tuples
        # Add gates from `circuit` to `newcircuit` (including measurements)
        for gate in circuit.queue:
            newcircuit.check_measured(gate.qubits)
            newcircuit.queue.append(gate)
            if isinstance(gate, gates.ParametrizedGate):
                newcircuit.parametrized_gates.append(gate)
                if gate.trainable:
                    newcircuit.trainable_gates.append(gate)

        if newcircuit.measurement_gate is None:
            newcircuit.measurement_gate = circuit.measurement_gate
            newcircuit.measurement_tuples = circuit.measurement_tuples
        elif circuit.measurement_gate is not None:
            for k, v in circuit.measurement_tuples.items():
                if k in newcircuit.measurement_tuples:
                    raise_error(KeyError, "Register name {} already exists in "
                                          "circuit.".format(k))
                newcircuit.check_measured(v)
                newcircuit.measurement_tuples[k] = v
            newcircuit.measurement_gate.add(circuit.measurement_gate)
        return newcircuit

    def on_qubits(self, *q):
        """Generator of gates contained in the circuit acting on specified qubits.

        Useful for adding a circuit as a subroutine in a larger circuit.

        Args:
            q (int): Qubit ids that the gates should act.

        Example:
            
            .. testcode::

                from qibo import gates, models
                # create small circuit on 4 qubits
                smallc = models.Circuit(4)
                smallc.add((gates.RX(i, theta=0.1) for i in range(4)))
                smallc.add((gates.CNOT(0, 1), gates.CNOT(2, 3)))
                # create large circuit on 8 qubits
                largec = models.Circuit(8)
                largec.add((gates.RY(i, theta=0.1) for i in range(8)))
                # add the small circuit to the even qubits of the large one
                largec.add(smallc.on_qubits(*range(0, 8, 2)))
        """
        if len(q) != self.nqubits:
            raise_error(ValueError, "Cannot return gates on {} qubits because "
                                    "the circuit contains {} qubits."
                                    "".format(len(q), self.nqubits))
        for gate in self.queue:
            yield gate._on_qubits(*q)

    def _shallow_copy(self):
        """Helper method for :meth:`qibo.abstractions.circuit.AbstractCircuit.copy`
        and :meth:`qibo.core.circuit.Circuit.fuse`."""
        new_circuit = self.__class__(**self.init_kwargs)
        new_circuit.parametrized_gates = _ParametrizedGates(self.parametrized_gates)
        new_circuit.trainable_gates = _ParametrizedGates(self.trainable_gates)
        new_circuit.measurement_gate = self.measurement_gate
        new_circuit.measurement_tuples = dict(self.measurement_tuples)
        return new_circuit

    def copy(self, deep: bool = False):
        """Creates a copy of the current ``circuit`` as a new ``Circuit`` model.

        Args:
            deep (bool): If ``True`` copies of the  gate objects will be created
                for the new circuit. If ``False``, the same gate objects of
                ``circuit`` will be used.

        Returns:
            The copied circuit object.
        """
        import copy
        if deep:
            new_circuit = self.__class__(**self.init_kwargs)
            for gate in self.queue:
                if isinstance(gate, gates.FusedGate): # pragma: no cover
                    # impractical case
                    raise_error(NotImplementedError, "Cannot create deep copy "
                                                     "of fused circuit.")

                new_gate = copy.copy(gate)
                new_circuit.queue.append(new_gate)
                if isinstance(gate, gates.ParametrizedGate):
                    new_circuit.parametrized_gates.append(new_gate)
                    if gate.trainable:
                        new_circuit.trainable_gates.append(new_gate)
            new_circuit.measurement_gate = copy.copy(self.measurement_gate)
            new_circuit.measurement_tuples = dict(self.measurement_tuples)
        else:
            new_circuit = self._shallow_copy()
            new_circuit.queue = copy.copy(self.queue)
        return new_circuit

    def invert(self):
        """Creates a new ``Circuit`` that is the inverse of the original.

        Inversion is obtained by taking the dagger of all gates in reverse order.
        If the original circuit contains measurement gates, these are included
        in the inverted circuit.

        Returns:
            The circuit inverse.
        """
        import copy
        new_circuit = self.__class__(**self.init_kwargs)
        for gate in self.queue[::-1]:
            new_circuit.add(gate.dagger())
        new_circuit.measurement_gate = copy.copy(self.measurement_gate)
        new_circuit.measurement_tuples = dict(self.measurement_tuples)
        return new_circuit

    def _check_noise_map(self, noise_map: NoiseMapType) -> NoiseMapType:
        if isinstance(noise_map, (tuple, list)):
            if len(noise_map) != 3:
                raise_error(ValueError, "Noise map expects three probabilities "
                                        "but received {}.".format(len(noise_map)))
            return {q: noise_map for q in range(self.nqubits)}
        elif isinstance(noise_map, dict):
            if len(noise_map) != self.nqubits:
                raise_error(ValueError, "Noise map has {} qubits while the circuit "
                                        "has {}.".format(len(noise_map), self.nqubits))
            for v in noise_map.values():
                if len(v) != 3:
                    raise_error(ValueError, "Noise map expects three probabilities "
                                            "but received {}.".format(v))
            return noise_map

        raise_error(TypeError, "Type {} of noise map is not recognized."
                               "".format(type(noise_map)))

    def decompose(self, *free: int):
        """Decomposes circuit's gates to gates supported by OpenQASM.

        Args:
            free: Ids of free (work) qubits to use for gate decomposition.

        Returns:
            Circuit that contains only gates that are supported by OpenQASM
            and has the same effect as the original circuit.
        """
        # FIXME: This method is not completed until the ``decompose`` is
        # implemented for all gates not supported by OpenQASM.
        decomp_circuit = self.__class__(self.nqubits)
        for gate in self.queue:
            decomp_circuit.add(gate.decompose(*free))
        decomp_circuit.measurement_tuples = dict(self.measurement_tuples)
        decomp_circuit.measurement_gate = self.measurement_gate
        return decomp_circuit

    def with_noise(self, noise_map: NoiseMapType):
        """Creates a copy of the circuit with noise gates after each gate.

        If the original circuit uses state vectors then noise simulation will
        be done using sampling and repeated circuit execution.
        In order to use density matrices the original circuit should be created
        using the ``density_matrix`` flag set to ``True``.
        For more information we refer to the
        :ref:`How to perform noisy simulation? <noisy-example>` example.

        Args:
            noise_map (dict): Dictionary that maps qubit ids to noise
                probabilities (px, py, pz).
                If a tuple of probabilities (px, py, pz) is given instead of
                a dictionary, then the same probabilities will be used for all
                qubits.

        Returns:
            Circuit object that contains all the gates of the original circuit
            and additional noise channels on all qubits after every gate.

        Example:
            .. testcode::

                from qibo.models import Circuit
                from qibo import gates
                # use density matrices for noise simulation
                c = Circuit(2, density_matrix=True)
                c.add([gates.H(0), gates.H(1), gates.CNOT(0, 1)])
                noise_map = {0: (0.1, 0.0, 0.2), 1: (0.0, 0.2, 0.1)}
                noisy_c = c.with_noise(noise_map)
                # ``noisy_c`` will be equivalent to the following circuit
                c2 = Circuit(2, density_matrix=True)
                c2.add(gates.H(0))
                c2.add(gates.PauliNoiseChannel(0, 0.1, 0.0, 0.2))
                c2.add(gates.H(1))
                c2.add(gates.PauliNoiseChannel(1, 0.0, 0.2, 0.1))
                c2.add(gates.CNOT(0, 1))
                c2.add(gates.PauliNoiseChannel(0, 0.1, 0.0, 0.2))
                c2.add(gates.PauliNoiseChannel(1, 0.0, 0.2, 0.1))
        """
        noise_map = self._check_noise_map(noise_map)
        # Generate noise gates
        noise_gates = []
        for gate in self.queue:
            if isinstance(gate, gates.KrausChannel):
                raise_error(ValueError, "`.with_noise` method is not available "
                                        "for circuits that already contain "
                                        "channels.")
            noise_gates.append([])
            for q in gate.qubits:
                if q in noise_map and sum(noise_map[q]) > 0:
                    p = noise_map[q]
                    noise_gates[-1].append(gate_module.PauliNoiseChannel(
                            q, px=p[0], py=p[1], pz=p[2]))

        # Create new circuit with noise gates inside
        noisy_circuit = self.__class__(self.nqubits)
        for i, gate in enumerate(self.queue):
            # Do not use `circuit.add` here because these gates are already
            # added in the original circuit
            noisy_circuit.queue.append(gate)
            for noise_gate in noise_gates[i]:
                noisy_circuit.add(noise_gate)
        noisy_circuit.parametrized_gates = list(self.parametrized_gates)
        noisy_circuit.trainable_gates = list(self.trainable_gates)
        noisy_circuit.measurement_tuples = dict(self.measurement_tuples)
        noisy_circuit.measurement_gate = self.measurement_gate
        return noisy_circuit

    def check_measured(self, gate_qubits: Tuple[int]):
        """Helper method for `add`.

        Checks if the qubits that a gate acts are already measured and raises
        a `NotImplementedError` if they are because currently we do not allow
        measured qubits to be reused.
        """
        for qubit in gate_qubits:
            if (self.measurement_gate is not None and
                qubit in self.measurement_gate.target_qubits):
                raise_error(ValueError, "Cannot reuse qubit {} because it is already "
                                        "measured".format(qubit))

    def add(self, gate):
        """Add a gate to a given queue.

        Args:
            gate (:class:`qibo.abstractions.gates.Gate`): the gate object to add.
                See :ref:`Gates` for a list of available gates.
                `gate` can also be an iterable or generator of gates.
                In this case all gates in the iterable will be added in the
                circuit.

        Returns:
            If the circuit contains measurement gates with ``collapse=True``
            a ``sympy.Symbol`` that parametrizes the corresponding outcome.
        """
        if isinstance(gate, collections.abc.Iterable):
            for g in gate:
                self.add(g)
        elif isinstance(gate, gates.Gate):
            return self._add(gate)
        else:
            raise_error(TypeError, "Unknown gate type {}.".format(type(gate)))

    def _add(self, gate: gates.Gate):
        if gate.density_matrix and not self.density_matrix:
            raise_error(ValueError, "Cannot add {} on circuits that uses state "
                                    "vectors. Please switch to density matrix "
                                    "circuit.".format(gate.name))
        elif self.density_matrix:
            gate.density_matrix = True

        if self._final_state is not None:
            raise_error(RuntimeError, "Cannot add gates to a circuit after it is "
                                      "executed.")

        for q in gate.target_qubits:
            if q >= self.nqubits:
                raise_error(ValueError, "Attempting to add gate with target qubits {} "
                                        "on a circuit of {} qubits."
                                        "".format(gate.target_qubits, self.nqubits))

        self.check_measured(gate.qubits)
        if isinstance(gate, gates.M) and not gate.collapse:
            self._add_measurement(gate)
        elif isinstance(gate, gates.VariationalLayer):
            self._add_layer(gate)
        else:
            self._set_nqubits(gate)
            self.queue.append(gate)
            if isinstance(gate, gates.M):
                self.repeated_execution = True
                return gate.symbol()
            if isinstance(gate, gates.UnitaryChannel):
                self.repeated_execution = not self.density_matrix
        if isinstance(gate, gates.ParametrizedGate):
            self.parametrized_gates.append(gate)
            if gate.trainable:
                self.trainable_gates.append(gate)

    def _set_nqubits(self, gate: gates.Gate):
        """Sets the number of qubits and prepares all gates.

        Helper method for ``circuit.add(gate)``.
        """
        if gate.is_prepared and gate.nqubits != self.nqubits:
            raise_error(RuntimeError, "Cannot add gate {} that acts on {} "
                                      "qubits to circuit that contains {}"
                                      "qubits.".format(
                                            gate, gate.nqubits, self.nqubits))

    def _add_measurement(self, gate: gates.Gate):
        """Called automatically by `add` when `gate` is measurement.

        This is because measurement gates (`gates.M`) are treated differently
        than all other gates.
        The user is not supposed to use the `_add_measurement` method.
        """
        # Set register's name and log the set of qubits in `self.measurement_tuples`
        name = gate.register_name
        if name is None:
            name = "register{}".format(len(self.measurement_tuples))
            gate.register_name = name
        elif name in self.measurement_tuples:
            raise_error(KeyError, "Register name {} has already been used."
                                  "".format(name))

        # Update circuit's global measurement gate
        if self.measurement_gate is None:
            self.measurement_gate = gate
            self.measurement_tuples[name] = tuple(gate.target_qubits)
        else:
            self.measurement_gate.add(gate)
            self.measurement_tuples[name] = gate.target_qubits

    @abstractmethod
    def _add_layer(self, gate: gates.Gate): # pragma: no cover
        """Called automatically by `add` when `gate` is of
        type `qibo.abstractions.gates.VariationalLayer`.
        """
        raise_error(NotImplementedError)

    @property
    def ngates(self) -> int:
        """Total number of gates/operations in the circuit."""
        return len(self.queue)

    @property
    def depth(self) -> int:
        """Circuit depth if each gate is placed at the earliest possible position."""
        return len(self.queue.moments)

    @property
    def gate_types(self) -> collections.Counter:
        """``collections.Counter`` with the number of appearances of each gate type.

        The QASM names are used as gate identifiers.
        """
        gatecounter = collections.Counter()
        for gate in self.queue:
            gatecounter[gate.name] += 1
        return gatecounter

    def gates_of_type(self, gate: Union[str, type]) -> List[Tuple[int, gates.Gate]]:
        """Finds all gate objects of specific type.

        Args:
            gate (str, type): The QASM name of a gate or the corresponding gate class.

        Returns:
            List with all gates that are in the circuit and have the same type
            with the given ``gate``. The list contains tuples ``(i, g)`` where
            ``i`` is the index of the gate ``g`` in the circuit's gate queue.
        """
        if isinstance(gate, str):
            return [(i, g) for i, g in enumerate(self.queue)
                    if g.name == gate]
        if isinstance(gate, type) and issubclass(gate, gates.Gate):
            return [(i, g) for i, g in enumerate(self.queue)
                    if isinstance(g, gate)]
        raise_error(TypeError, "Gate identifier {} not recognized.".format(gate))

    def _set_parameters_list(self, parameters: List, n: int):
        """Helper method for ``set_parameters`` when a list is given.

        Also works if ``parameters`` is ``np.ndarray`` or ``tf.Tensor``.
        """
        if n == len(self.trainable_gates):
            for i, gate in enumerate(self.trainable_gates):
                gate.parameters = parameters[i]
        elif n == self.trainable_gates.nparams:
            parameters = list(parameters)
            k = 0
            for i, gate in enumerate(self.trainable_gates):
                if gate.nparams == 1:
                    gate.parameters = parameters[i + k]
                else:
                    gate.parameters = parameters[i + k: i + k + gate.nparams]
                k += gate.nparams - 1
        else:
            raise_error(ValueError, "Given list of parameters has length {} while "
                                    "the circuit contains {} parametrized gates."
                                    "".format(n, len(self.trainable_gates)))

    def set_parameters(self, parameters):
        """Updates the parameters of the circuit's parametrized gates.

        For more information on how to use this method we refer to the
        :ref:`How to use parametrized gates?<params-examples>` example.

        Args:
            parameters: Container holding the new parameter values.
                It can have one of the following types:
                List with length equal to the number of parametrized gates and
                each of its elements compatible with the corresponding gate.
                Dictionary with keys that are references to the parametrized
                gates and values that correspond to the new parameters for
                each gate.
                Flat list with length equal to the total number of free
                parameters in the circuit.
                A backend supported tensor (for example ``np.ndarray`` or
                ``tf.Tensor``) may also be given instead of a flat list.


        Example:
            .. testcode::

                from qibo.models import Circuit
                from qibo import gates
                # create a circuit with all parameters set to 0.
                c = Circuit(3)
                c.add(gates.RX(0, theta=0))
                c.add(gates.RY(1, theta=0))
                c.add(gates.CZ(1, 2))
                c.add(gates.fSim(0, 2, theta=0, phi=0))
                c.add(gates.H(2))

                # set new values to the circuit's parameters using list
                params = [0.123, 0.456, (0.789, 0.321)]
                c.set_parameters(params)
                # or using dictionary
                params = {c.queue[0]: 0.123, c.queue[1]: 0.456,
                          c.queue[3]: (0.789, 0.321)}
                c.set_parameters(params)
                # or using flat list (or an equivalent `np.array`/`tf.Tensor`)
                params = [0.123, 0.456, 0.789, 0.321]
                c.set_parameters(params)
        """
        if isinstance(parameters, (list, tuple)):
            self._set_parameters_list(parameters, len(parameters))
        elif isinstance(parameters, self.param_tensor_types):
            self._set_parameters_list(parameters, int(parameters.shape[0]))
        elif isinstance(parameters, dict):
            diff = set(parameters.keys()) - self.trainable_gates.set
            if diff:
                raise_error(KeyError, "Dictionary contains gates {} which are "
                                      "not on the list of parametrized gates "
                                      "of the circuit.".format(diff))
            for gate, params in parameters.items():
                gate.parameters = params
        else:
            raise_error(TypeError, "Invalid type of parameters {}."
                                   "".format(type(parameters)))
        # Reset ``FusedGate`` matrices so that they are recalculated with the
        # updated parameters.
        for gate in self.queue:
            if isinstance(gate, gates.FusedGate):
                gate._reset_unitary()

    def get_parameters(self, format: str = "list",
                       include_not_trainable: bool = False
                       ) -> Union[List, Dict]: # pylint: disable=W0622
        """Returns the parameters of all parametrized gates in the circuit.

        Inverse method of :meth:`qibo.abstractions.circuit.AbstractCircuit.set_parameters`.

        Args:
            format (str): How to return the variational parameters.
                Available formats are ``'list'``, ``'dict'`` and ``'flatlist'``.
                See :meth:`qibo.abstractions.circuit.AbstractCircuit.set_parameters`
                for more details on each format. Default is ``'list'``.
            include_not_trainable (bool): If ``True`` it includes the parameters
                of non-trainable parametrized gates in the returned list or
                dictionary. Default is ``False``.
        """
        if include_not_trainable:
            parametrized_gates = self.parametrized_gates
        else:
            parametrized_gates = self.trainable_gates

        if format == "list":
            params = [gate.parameters for gate in parametrized_gates]
        elif format == "dict":
            params = {gate: gate.parameters for gate in parametrized_gates}
        elif format == "flatlist":
            params = []
            for gate in parametrized_gates:
                if isinstance(gate.parameters, self.param_tensor_types):
                    def traverse(x):
                        if isinstance(x, self.param_tensor_types):
                            for v1 in x:
                                for v2 in traverse(v1):
                                    yield v2
                        else:
                            yield x
                    params.extend(traverse(gate.parameters))
                elif isinstance(gate.parameters, collections.abc.Iterable):
                    params.extend(gate.parameters)
                else:
                    params.append(gate.parameters)
        else:
            raise_error(ValueError, "Unknown format {} given in "
                                    "``get_parameters``.".format(format))
        return params

    def summary(self) -> str:
        """Generates a summary of the circuit.

        The summary contains the circuit depths, total number of qubits and
        the all gates sorted in decreasing number of appearance.

        Example:
            .. testcode::

                from qibo.models import Circuit
                from qibo import gates
                c = Circuit(3)
                c.add(gates.H(0))
                c.add(gates.H(1))
                c.add(gates.CNOT(0, 2))
                c.add(gates.CNOT(1, 2))
                c.add(gates.H(2))
                c.add(gates.TOFFOLI(0, 1, 2))
                print(c.summary())
                # Prints
                '''
                Circuit depth = 5
                Total number of gates = 6
                Number of qubits = 3
                Most common gates:
                h: 3
                cx: 2
                ccx: 1
                '''
        
            .. testoutput::
                :hide:

                Circuit depth = 5
                Total number of gates = 6
                Number of qubits = 3
                Most common gates:
                h: 3
                cx: 2
                ccx: 1
        """
        logs = [f"Circuit depth = {self.depth}",
                f"Total number of gates = {self.ngates}",
                f"Number of qubits = {self.nqubits}",
                "Most common gates:"]
        common_gates = self.gate_types.most_common()
        logs.extend("{}: {}".format(g, n) for g, n in common_gates)
        return "\n".join(logs)

    @abstractmethod
    def fuse(self): # pragma: no cover
        raise_error(NotImplementedError)

    @property
    @abstractmethod
    def final_state(self): # pragma: no cover
        """Returns the final state after full simulation of the circuit.

        If the circuit is executed more than once, only the last final state
        is returned.
        """
        raise_error(NotImplementedError)

    @abstractmethod
    def execute(self, initial_state=None, nshots=None): # pragma: no cover
        """Executes the circuit. Exact implementation depends on the backend.

        See :meth:`qibo.core.circuit.Circuit.execute` for more
        details.
        """
        raise_error(NotImplementedError)

    def __call__(self, initial_state=None, nshots=None):
        """Equivalent to ``circuit.execute``."""
        return self.execute(initial_state=initial_state, nshots=nshots)

    def to_qasm(self):
        """Convert circuit to QASM.

        Args:
            filename (str): The filename where the code is saved.
        """
        from qibo import __version__
        code = [f"// Generated by QIBO {__version__}"]
        code += ["OPENQASM 2.0;"]
        code += ["include \"qelib1.inc\";"]
        code += [f"qreg q[{self.nqubits}];"]

        # Set measurements
        for reg_name, reg_qubits in self.measurement_tuples.items():
            if not reg_name.islower():
                raise_error(NameError, "OpenQASM does not support capital letters in "
                                       "register names but {} was used".format(reg_name))
            code.append(f"creg {reg_name}[{len(reg_qubits)}];")

        # Add gates
        for gate in self.queue:
            if gate.name not in gates.QASM_GATES:
                raise_error(ValueError, f"Gate {gate.name} is not supported by OpenQASM.")
            if gate.is_controlled_by:
                raise_error(ValueError, "OpenQASM does not support multi-controlled gates.")

            qubits = ",".join(f"q[{i}]" for i in gate.qubits)
            if gate.name in gates.PARAMETRIZED_GATES:
                if isinstance(gate.parameters, collections.abc.Iterable):
                    params = (str(x) for x in gate.parameters)
                    name = "{}({})".format(gate.name, ", ".join(params))
                else:
                    name = f"{gate.name}({gate.parameters})"
            else:
                name = gate.name
            code.append(f"{name} {qubits};")

        # Add measurements
        for reg_name, reg_qubits in self.measurement_tuples.items():
            for i, q in enumerate(reg_qubits):
                code.append(f"measure q[{q}] -> {reg_name}[{i}];")

        return "\n".join(code)

    @classmethod
    def from_qasm(cls, qasm_code: str, **kwargs):
        """Constructs a circuit from QASM code.

        Args:
            qasm_code (str): String with the QASM script.

        Returns:
            A :class:`qibo.abstractions.circuit.AbstractCircuit` that contains the gates
            specified by the given QASM script.

        Example:
         
            .. testcode::

                from qibo import models, gates
                qasm_code = '''OPENQASM 2.0;
                include "qelib1.inc";
                qreg q[2];
                h q[0];
                h q[1];
                cx q[0],q[1];'''
                c = models.Circuit.from_qasm(qasm_code)
                # is equivalent to creating the following circuit
                c2 = models.Circuit(2)
                c2.add(gates.H(0))
                c2.add(gates.H(1))
                c2.add(gates.CNOT(0, 1))
        """
        kwargs["nqubits"], gate_list = cls._parse_qasm(qasm_code)
        circuit = cls(**kwargs)
        for gate_name, qubits, params in gate_list:
            gate = getattr(gate_module, gate_name)
            if gate_name == "M":
                circuit.add(gate(*qubits, register_name=params))
            elif params is None:
                circuit.add(gate(*qubits))
            else:
                # assume parametrized gate
                circuit.add(gate(*qubits, *params))
        return circuit

    @staticmethod
    def _parse_qasm(qasm_code: str
                    ) -> Tuple[int,
                               List[Tuple[str, List[int],
                                          Optional[Union[str, float]]]]]:
        """Extracts circuit information from QASM script.

        Helper method for ``from_qasm``.

        Args:
            qasm_code: String with the QASM code to parse.

        Returns:
            nqubits: The total number of qubits in the circuit.
            gate_list: List that specifies the gates of the circuit.
                Contains tuples of the form
                (Qibo gate name, qubit IDs, optional additional parameter).
                The additional parameter is the ``register_name`` for
                measurement gates or ``theta`` for parametrized gates.
        """
        import re
        def read_args(args):
            _args = iter(re.split(r"[\[\],]", args))
            for name in _args:
                if name:
                    index = next(_args)
                    if not index.isdigit():
                        raise_error(ValueError, "Invalid QASM qubit arguments:", args)
                    yield name, int(index)

        # Remove comment lines
        lines = "".join(line for line in qasm_code.split("\n")
                        if line and line[:2] != "//")
        lines = (line for line in lines.split(";") if line)

        if next(lines) != "OPENQASM 2.0":
            raise_error(ValueError, "QASM code should start with 'OPENQASM 2.0'.")

        qubits = {} # Dict[Tuple[str, int], int]: map from qubit tuple to qubit id
        cregs_size = {} # Dict[str, int]: map from `creg` name to its size
        registers = {} # Dict[str, List[int]]: map from register names to target qubit ids
        gate_list = [] # List[Tuple[str, List[int]]]: List of (gate name, list of target qubit ids)
        for line in lines:
            command, args = line.split(None, 1)
            # remove spaces
            command = command.replace(" ", "")
            args = args.replace(" ", "")

            if command == "include":
                pass

            elif command == "qreg":
                for name, nqubits in read_args(args):
                    for i in range(nqubits):
                        qubits[(name, i)] = len(qubits)

            elif command == "creg":
                for name, nqubits in read_args(args):
                    cregs_size[name] = nqubits

            elif command == "measure":
                args = args.split("->")
                if len(args) != 2:
                    raise_error(ValueError, "Invalid QASM measurement:", line)
                qubit = next(read_args(args[0]))
                if qubit not in qubits:
                    raise_error(ValueError, "Qubit {} is not defined in QASM code."
                                            "".format(qubit))

                register, idx = next(read_args(args[1]))
                if register not in cregs_size:
                    raise_error(ValueError, "Classical register name {} is not defined "
                                            "in QASM code.".format(register))
                if idx >= cregs_size[register]:
                    raise_error(ValueError, "Cannot access index {} of register {} "
                                            "with {} qubits."
                                            "".format(idx, register, cregs_size[register]))
                if register in registers:
                    if idx in registers[register]:
                        raise_error(KeyError, "Key {} of register {} has already "
                                              "been used.".format(idx, register))
                    registers[register][idx] = qubits[qubit]
                else:
                    registers[register] = {idx: qubits[qubit]}
                    gate_list.append(("M", register, None))

            else:
                pieces = [x for x in re.split("[()]", command) if x]
                if len(pieces) == 1:
                    gatename, params = pieces[0], None
                    if gatename not in gates.QASM_GATES:
                        raise_error(ValueError, "QASM command {} is not recognized."
                                                "".format(command))
                    if gatename in gates.PARAMETRIZED_GATES:
                        raise_error(ValueError, "Missing parameters for QASM "
                                                "gate {}.".format(gatename))

                elif len(pieces) == 2:
                    gatename, params = pieces
                    if gatename not in gates.PARAMETRIZED_GATES:
                        raise_error(ValueError, "Invalid QASM command {}."
                                                "".format(command))
                    params = params.replace(" ", "").split(",")
                    try:
                        for i, p in enumerate(params):
                            if 'pi' in p:
                                import math
                                from operator import mul
                                from functools import reduce
                                s = p.replace('pi', str(math.pi)).split('*')
                                p = reduce(mul, [float(j) for j in s], 1)
                            params[i] = float(p)
                    except ValueError:
                        raise_error(ValueError, "Invalid value {} for gate parameters."
                                                "".format(params))

                else:
                    raise_error(ValueError, "QASM command {} is not recognized."
                                            "".format(command))

                # Add gate to gate list
                qubit_list = []
                for qubit in read_args(args):
                    if qubit not in qubits:
                        raise_error(ValueError, "Qubit {} is not defined in QASM "
                                                "code.".format(qubit))
                    qubit_list.append(qubits[qubit])
                gate_list.append((gates.QASM_GATES[gatename],
                                  list(qubit_list),
                                  params))

        # Create measurement gate qubit lists from registers
        for i, (gatename, register, _) in enumerate(gate_list):
            if gatename == "M":
                qubit_list = registers[register]
                qubit_list = [qubit_list[k] for k in sorted(qubit_list.keys())]
                gate_list[i] = ("M", qubit_list, register)

        return len(qubits), gate_list

    def draw(self, line_wrap=70) -> str:
        """Draw text circuit using unicode symbols.

        Args:
            line_wrap (int): maximum number of characters per line. This option
                split the circuit text diagram in chunks of line_wrap characters.

        Return:
            String containing text circuit diagram.
        """
        labels = {"h": "H", "x": "X", "y": "Y", "z": "Z",
                  "rx": "RX", "ry": "RY", "rz": "RZ",
                  "u1": "U1", "u2": "U2", "u3": "U3",
                  "cx": "X", "swap": "x", "cz": "Z",
                  "crx": "RX", "cry": "RY", "crz": "RZ",
                  "cu1": "U1", "cu3": "U3", "ccx": "X",
                  "id": "I", "measure": "M", "fsim": "f",
                  "generalizedfsim": "gf", "Unitary": "U"}

        # build string representation of gates
        matrix = [[] for _ in range(self.nqubits)]
        idx = [0] * self.nqubits

        for gate in self.queue:
            if gate.name not in labels:
                raise_error(NotImplementedError, f"{gate.name} gate is not supported by `circuit.draw`")
            gate_name = labels.get(gate.name)
            targets = list(gate.target_qubits)
            controls = list(gate.control_qubits)

            # identify boundaries
            qubits = targets + controls
            qubits.sort()
            min_qubits_id = qubits[0]
            max_qubits_id = qubits[-1]

            # identify column
            col = idx[targets[0]] if not controls and len(targets) == 1 else max(idx)

            # extend matrix
            for iq in range(self.nqubits):
                matrix[iq].extend((1 + col - len(matrix[iq]))* [''])

            # fill
            for iq in range(min_qubits_id, max_qubits_id + 1):
                if iq in targets:
                    matrix[iq][col] = gate_name
                elif iq in controls:
                    matrix[iq][col] = 'o'
                else:
                    matrix[iq][col] = '|'

            # update indexes
            if not controls and len(targets) == 1:
                idx[targets[0]] += 1
            else:
                idx = [col + 1] * self.nqubits

        # Include measurement gates
        if self.measurement_gate:
            for iq in range(self.nqubits):
                matrix[iq].append('M' if iq in self.measurement_gate.target_qubits else '')

        # Add some spacers
        for col in range(len(matrix[0])):
            maxlen = max([len(matrix[l][col]) for l in range(self.nqubits)])
            for row in range(self.nqubits):
                matrix[row][col] += '─' * (1 + maxlen - len(matrix[row][col]))

        # Print to terminal
        output = ""
        for q in range(self.nqubits):
            output += f'q{q}' + ' ' * (len(str(self.nqubits))-len(str(q))) + \
                       ': ─' + ''.join(matrix[q]) + '\n'

        # line wrap
        if line_wrap:
            loutput = output.splitlines()
            def chunkstring(string, length):
                nchunks = range(0, len(string), length)
                return (string[i:length+i] for i in nchunks), len(nchunks)
            for row in range(self.nqubits):
                chunks, nchunks = chunkstring(loutput[row][3 + len(str(self.nqubits)):], line_wrap)
                if nchunks == 1:
                    loutput = None
                    break
                for i, c in enumerate(chunks):
                    loutput += ['' for _ in range(self.nqubits)]
                    suffix = ' ...\n'
                    prefix = f'q{row}' + ' ' * (len(str(self.nqubits))-len(str(row))) + ': '
                    if i == 0:
                        prefix += ' ' * 4
                    elif row == 0:
                        prefix = '\n' + prefix + '... '
                    else:
                        prefix += '... '
                    if i == nchunks-1:
                        suffix = '\n'
                    loutput[row + i * self.nqubits] = prefix + c + suffix
            if loutput is not None:
                output = ''.join(loutput)

        return output.rstrip('\n')



"""Test functions defined in `qibo/core/distcircuit.py`."""
import pytest
import numpy as np
from qibo import K, gates
from qibo.models import Circuit
from qibo.core.distcircuit import DistributedCircuit
from qibo.tests.utils import random_state


@pytest.mark.parametrize("use_global_qubits", [False, True])
def test_distributed_circuit_execution(backend, accelerators, use_global_qubits):
    dist_c = DistributedCircuit(6, accelerators)
    c = Circuit(6)
    if use_global_qubits:
        dist_c.add((gates.H(i) for i in range(dist_c.nqubits)))
        c.add((gates.H(i) for i in range(dist_c.nqubits)))
    else:
        dist_c.add((gates.H(i) for i in range(dist_c.nlocal)))
        c.add((gates.H(i) for i in range(dist_c.nlocal)))
    dist_c.global_qubits = range(dist_c.nlocal, dist_c.nqubits)

    initial_state = random_state(c.nqubits)
    final_state = dist_c(np.copy(initial_state))
    target_state = c(np.copy(initial_state))
    K.assert_allclose(target_state, final_state)


def test_distributed_circuit_execution_pretransformed(backend, accelerators):
    dist_c = DistributedCircuit(4, accelerators)
    dist_c.add((gates.H(i) for i in range(dist_c.nglobal, 4)))
    dist_c.add(gates.SWAP(0, 2))
    dist_c.add((gates.H(i) for i in range(dist_c.nglobal, 4)))

    c = Circuit(4)
    c.add((gates.H(i) for i in range(dist_c.nglobal, 4)))
    c.add(gates.SWAP(0, 2))
    c.add((gates.H(i) for i in range(dist_c.nglobal, 4)))

    initial_state = random_state(c.nqubits)
    final_state = dist_c(np.copy(initial_state))
    target_state = c(np.copy(initial_state))
    K.assert_allclose(target_state, final_state)


def test_distributed_circuit_execution_with_swap(backend, accelerators):
    dist_c = DistributedCircuit(6, accelerators)
    dist_c.add((gates.H(i) for i in range(6)))
    dist_c.add((gates.SWAP(i, i + 1) for i in range(5)))
    dist_c.global_qubits = [0, 1]

    c = Circuit(6)
    c.add((gates.H(i) for i in range(6)))
    c.add((gates.SWAP(i, i + 1) for i in range(5)))

    initial_state = random_state(c.nqubits)
    final_state = dist_c(np.copy(initial_state))
    target_state = c(np.copy(initial_state))
    K.assert_allclose(target_state, final_state)


def test_distributed_circuit_execution_special_gate(backend, accelerators):
    dist_c = DistributedCircuit(6, accelerators)
    initial_state = random_state(dist_c.nqubits)
    dist_c.add(gates.Flatten(np.copy(initial_state)))
    dist_c.add((gates.H(i) for i in range(dist_c.nlocal)))
    dist_c.global_qubits = range(dist_c.nlocal, dist_c.nqubits)
    c = Circuit(6)
    c.add(gates.Flatten(np.copy(initial_state)))
    c.add((gates.H(i) for i in range(dist_c.nlocal)))
    K.assert_allclose(dist_c(), c())


def test_distributed_circuit_execution_controlled_gate(backend, accelerators):
    dist_c = DistributedCircuit(4, accelerators)
    dist_c.add((gates.H(i) for i in range(dist_c.nglobal, 4)))
    dist_c.add(gates.CNOT(0, 2))

    c = Circuit(4)
    c.add((gates.H(i) for i in range(dist_c.nglobal, 4)))
    c.add(gates.CNOT(0, 2))
    initial_state = random_state(c.nqubits)
    final_state = dist_c(np.copy(initial_state))
    target_state = c(np.copy(initial_state))
    K.assert_allclose(target_state, final_state)


def test_distributed_circuit_execution_controlled_by_gates(backend, accelerators):
    dist_c = DistributedCircuit(6, accelerators)
    dist_c.add([gates.H(0), gates.H(2), gates.H(3)])
    dist_c.add(gates.CNOT(4, 5))
    dist_c.add(gates.Z(1).controlled_by(0))
    dist_c.add(gates.SWAP(2, 3))
    dist_c.add([gates.X(2), gates.X(3), gates.X(4)])

    c = Circuit(6)
    c.add([gates.H(0), gates.H(2), gates.H(3)])
    c.add(gates.CNOT(4, 5))
    c.add(gates.Z(1).controlled_by(0))
    c.add(gates.SWAP(2, 3))
    c.add([gates.X(2), gates.X(3), gates.X(4)])

    initial_state = random_state(c.nqubits)
    final_state = dist_c(np.copy(initial_state))
    target_state = c(np.copy(initial_state))
    K.assert_allclose(target_state, final_state)


def test_distributed_circuit_execution_addition(backend, accelerators):
    # Attempt to add circuits with different devices
    c1 = DistributedCircuit(6, {"/GPU:0": 2, "/GPU:1": 2})
    c2 = DistributedCircuit(6, {"/GPU:0": 2})
    with pytest.raises(ValueError):
        c = c1 + c2

    c1 = DistributedCircuit(6, accelerators)
    c2 = DistributedCircuit(6, accelerators)
    c1.add([gates.H(i) for i in range(6)])
    c2.add([gates.CNOT(i, i + 1) for i in range(5)])
    c2.add([gates.Z(i) for i in range(6)])
    dist_c = c1 + c2

    c = Circuit(6)
    c.add([gates.H(i) for i in range(6)])
    c.add([gates.CNOT(i, i + 1) for i in range(5)])
    c.add([gates.Z(i) for i in range(6)])
    assert c.depth == dist_c.depth
    K.assert_allclose(dist_c(), c())


def test_distributed_circuit_empty_execution(backend, accelerators):
    # test executing a circuit with the default initial state
    c = DistributedCircuit(5, accelerators)
    final_state = c().state()
    target_state = np.zeros_like(final_state)
    target_state[0] = 1
    K.assert_allclose(final_state, target_state)
    # test re-executing the circuit with a given initial state
    initial_state = random_state(c.nqubits)
    K.assert_allclose(c(initial_state), initial_state)
    # test executing a new circuit with a given initial state
    c = DistributedCircuit(5, accelerators)
    initial_state = random_state(c.nqubits)
    K.assert_allclose(c(initial_state), initial_state)
    # test re-executing the circuit with the default initial state
    K.assert_allclose(c(), target_state)



"""Test :class:`qibo.abstractions.gates.M` when results are probabilistic."""
import sys
import pytest
import numpy as np
import qibo
from qibo import models, gates, K
from qibo.tests.test_measurement_gate import assert_result


@pytest.mark.parametrize("use_samples", [True, False])
def test_probabilistic_measurement(backend, accelerators, use_samples):
    original_threads = qibo.get_threads()
    # set single-thread to fix the random values generated from the frequency custom op
    qibo.set_threads(1)
    c = models.Circuit(4, accelerators)
    c.add(gates.H(0))
    c.add(gates.H(1))
    c.add(gates.M(0, 1))
    result = c(nshots=1000)

    K.set_seed(1234)
    if use_samples:
        # calculates sample tensor directly using `tf.random.categorical`
        # otherwise it uses the frequency-only calculation
        _ = result.samples()

    # update reference values based on backend and device
    decimal_frequencies = K.test_regressions("test_probabilistic_measurement")
    assert sum(result.frequencies().values()) == 1000
    assert_result(result, decimal_frequencies=decimal_frequencies)
    qibo.set_threads(original_threads)


@pytest.mark.parametrize("use_samples", [True, False])
def test_unbalanced_probabilistic_measurement(backend, use_samples):
    original_threads = qibo.get_threads()
    # set single-thread to fix the random values generated from the frequency custom op
    qibo.set_threads(1)
    state = np.array([1, 1, 1, np.sqrt(3)]) / np.sqrt(6)
    c = models.Circuit(2)
    c.add(gates.Flatten(state))
    c.add(gates.M(0, 1))
    result = c(nshots=1000)

    K.set_seed(1234)
    if use_samples:
        # calculates sample tensor directly using `tf.random.categorical`
        # otherwise it uses the frequency-only calculation
        _ = result.samples()
    # update reference values based on backend and device
    decimal_frequencies = K.test_regressions("test_unbalanced_probabilistic_measurement")
    assert sum(result.frequencies().values()) == 1000
    assert_result(result, decimal_frequencies=decimal_frequencies)
    qibo.set_threads(original_threads)


def test_measurements_with_probabilistic_noise(backend):
    """Check measurements when simulating noise with repeated execution."""
    thetas = np.random.random(5)
    c = models.Circuit(5)
    c.add((gates.RX(i, t) for i, t in enumerate(thetas)))
    c.add((gates.PauliNoiseChannel(i, px=0.0, py=0.2, pz=0.4, seed=123)
           for i in range(5)))
    c.add(gates.M(*range(5)))
    K.set_seed(123)
    samples = c(nshots=20).samples()

    np.random.seed(123)
    K.set_seed(123)
    target_samples = []
    for _ in range(20):
        noiseless_c = models.Circuit(5)
        noiseless_c.add((gates.RX(i, t) for i, t in enumerate(thetas)))
        for i in range(5):
            if np.random.random() < 0.2:
                noiseless_c.add(gates.Y(i))
            if np.random.random() < 0.4:
                noiseless_c.add(gates.Z(i))
        noiseless_c.add(gates.M(*range(5)))
        target_samples.append(noiseless_c(nshots=1).samples())
    target_samples = np.concatenate(target_samples, axis=0)
    K.assert_allclose(samples, target_samples)


@pytest.mark.parametrize("i,probs", [(0, [0.0, 0.0, 0.0]),
                                     (1, [0.1, 0.3, 0.2]),
                                     (2, [0.5, 0.5, 0.5])])
def test_post_measurement_bitflips_on_circuit(backend, accelerators, i, probs):
    """Check bitflip errors on circuit measurements."""
    K.set_seed(123)
    c = models.Circuit(5, accelerators=accelerators)
    c.add([gates.X(0), gates.X(2), gates.X(3)])
    c.add(gates.M(0, 1, p0={0: probs[0], 1: probs[1]}))
    c.add(gates.M(3, p0=probs[2]))
    result = c(nshots=30).frequencies(binary=False)
    targets = K.test_regressions("test_post_measurement_bitflips_on_circuit")
    assert result == targets[i]


def test_post_measurement_bitflips_on_circuit_result(backend):
    """Check bitflip errors on ``CircuitResult`` objects."""
    thetas = np.random.random(4)
    c = models.Circuit(4)
    c.add((gates.RX(i, theta=t) for i, t in enumerate(thetas)))
    c.add(gates.M(0, 1, register_name="a"))
    c.add(gates.M(3, register_name="b"))
    result = c(nshots=30)
    samples = result.samples()
    K.set_seed(123)
    noisy_result = result.copy().apply_bitflips({0: 0.2, 1: 0.4, 3: 0.3})
    noisy_samples = noisy_result.samples(binary=True)
    register_samples = noisy_result.samples(binary=True, registers=True)

    K.set_seed(123)
    sprobs = K.to_numpy(K.random_uniform(samples.shape))
    flipper = K.cast(sprobs < np.array([0.2, 0.4, 0.3]), dtype=samples.dtype)
    target_samples = (samples + flipper) % 2
    K.assert_allclose(noisy_samples, target_samples)
    # Check register samples
    K.assert_allclose(register_samples["a"], target_samples[:, :2])
    K.assert_allclose(register_samples["b"], target_samples[:, 2:])
